/********************************************************************************************************************************/      
/* Disclaimer
/* This code is provided as is, without any guarantees.
/* Comments and bug reports should be forwarded to bilge@rad.upenn.edu.
/* DeformationJacobian3D -- Copyright Bilge Karacali 2005
/********************************************************************************************************************************/      

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <dfieldio.h>

/********************************************************************************************************************************/	

/* Predefined constants */
#define PI 3.14159265358979

/********************************************************************************************************************************/	

/* Global Variables */
int R=256;
int C=256;
int S=198;
char Deformationfilename[600];
char Outputfilename[600];
char *Defaultoutputfilename="defaultoutput.dat";
int NODEFORMATION=1;
int NOOUTFILE=1;
int DISP=1;
float iDX=1.0;
float iDY=1.0;
float iDZ=1.0;
int DOUBLEOUTPUT=0;

/********************************************************************************************************************************/	

/* Function Declarations */
void Usage();
int HandleOptions();
int DeformationJacobian3D(float *, float *, float *, double **, int, int, int, int, float, float, float);

/********************************************************************************************************************************/	
/* Function Definitions */

int DeformationJacobian3D(float *f,
                          float *g,
                          float *e,
                          double **J,
                          int r,
                          int c,
                          int s,
                          int disp,
                          float idx,
                          float idy,
                          float idz)
{
    int i,j,k,nbind,rc=r*c,rcs=rc*s;
    double cJ;
    int ccc,ccp,cpc,pcc,ccm,cmc,mcc;
    double fx000, fy000, fz000, gx000, gy000, gz000, ex000, ey000, ez000;
    double fx001, fy001, fz001, gx001, gy001, gz001, ex001, ey001, ez001;
    double fx010, fy010, fz010, gx010, gy010, gz010, ex010, ey010, ez010;
    double fx011, fy011, fz011, gx011, gy011, gz011, ex011, ey011, ez011;
    double fx100, fy100, fz100, gx100, gy100, gz100, ex100, ey100, ez100;
    double fx101, fy101, fz101, gx101, gy101, gz101, ex101, ey101, ez101;
    double fx110, fy110, fz110, gx110, gy110, gz110, ex110, ey110, ez110;
    double fx111, fy111, fz111, gx111, gy111, gz111, ex111, ey111, ez111;
    float f000,f001,f010,f011,f100,f101,f110,f111;
    float g000,g001,g010,g011,g100,g101,g110,g111;
    float e000,e001,e010,e011,e100,e101,e110,e111;
    float J000,J001,J010,J011,J100,J101,J110,J111;
    int i000,i001,i010,i011,i100,i101,i110,i111;
    int *NB;
    
    NB=(int *)calloc(rcs,sizeof(int));
    
    for (k=0;k<s-1;k++)
        for (j=0;j<c-1;j++)
            for (i=0;i<r-1;i++)
            {
                i000=k*rc+j*r+i;
                i001=i000+r;
                i010=i000+1;
                i011=i001+1;
                i100=i000+rc;
                i101=i100+r;
                i110=i100+1;
                i111=i101+1;
                
                f000=f[i000]; f001=f[i001]; f010=f[i010]; f011=f[i011]; f100=f[i100]; f101=f[i101]; f110=f[i110]; f111=f[i111]; 
                g000=g[i000]; g001=g[i001]; g010=g[i010]; g011=g[i011]; g100=g[i100]; g101=g[i101]; g110=g[i110]; g111=g[i111]; 
                e000=e[i000]; e001=e[i001]; e010=e[i010]; e011=e[i011]; e100=e[i100]; e101=e[i101]; e110=e[i110]; e111=e[i111]; 
                
                fx000=(double)(f001-f000); fy000=(double)(f010-f000); fz000=(double)(f100-f000); fx001=(double)(f001-f000); fy001=(double)(f011-f001); fz001=(double)(f101-f001); fx010=(double)(f011-f010); fy010=(double)(f010-f000); fz010=(double)(f110-f010); fx011=(double)(f011-f010); fy011=(double)(f011-f001); fz011=(double)(f111-f011); fx100=(double)(f101-f100); fy100=(double)(f110-f100); fz100=(double)(f100-f000); fx101=(double)(f101-f100); fy101=(double)(f111-f101); fz101=(double)(f101-f001); fx110=(double)(f111-f110); fy110=(double)(f110-f100); fz110=(double)(f110-f010); fx111=(double)(f111-f110); fy111=(double)(f111-f101); fz111=(double)(f111-f011); 
                gx000=(double)(g001-g000); gy000=(double)(g010-g000); gz000=(double)(g100-g000); gx001=(double)(g001-g000); gy001=(double)(g011-g001); gz001=(double)(g101-g001); gx010=(double)(g011-g010); gy010=(double)(g010-g000); gz010=(double)(g110-g010); gx011=(double)(g011-g010); gy011=(double)(g011-g001); gz011=(double)(g111-g011); gx100=(double)(g101-g100); gy100=(double)(g110-g100); gz100=(double)(g100-g000); gx101=(double)(g101-g100); gy101=(double)(g111-g101); gz101=(double)(g101-g001); gx110=(double)(g111-g110); gy110=(double)(g110-g100); gz110=(double)(g110-g010); gx111=(double)(g111-g110); gy111=(double)(g111-g101); gz111=(double)(g111-g011); 
                ex000=(double)(e001-e000); ey000=(double)(e010-e000); ez000=(double)(e100-e000); ex001=(double)(e001-e000); ey001=(double)(e011-e001); ez001=(double)(e101-e001); ex010=(double)(e011-e010); ey010=(double)(e010-e000); ez010=(double)(e110-e010); ex011=(double)(e011-e010); ey011=(double)(e011-e001); ez011=(double)(e111-e011); ex100=(double)(e101-e100); ey100=(double)(e110-e100); ez100=(double)(e100-e000); ex101=(double)(e101-e100); ey101=(double)(e111-e101); ez101=(double)(e101-e001); ex110=(double)(e111-e110); ey110=(double)(e110-e100); ez110=(double)(e110-e010); ex111=(double)(e111-e110); ey111=(double)(e111-e101); ez111=(double)(e111-e011); 
                
                if (idx!=1.0)
                {
                    fx000*=(double)idx; gx000*=(double)idx; ex000*=(double)idx;
                    fx001*=(double)idx; gx001*=(double)idx; ex001*=(double)idx;
                    fx010*=(double)idx; gx010*=(double)idx; ex010*=(double)idx;
                    fx011*=(double)idx; gx011*=(double)idx; ex011*=(double)idx;
                    fx100*=(double)idx; gx100*=(double)idx; ex100*=(double)idx;
                    fx101*=(double)idx; gx101*=(double)idx; ex101*=(double)idx;
                    fx110*=(double)idx; gx110*=(double)idx; ex110*=(double)idx;
                    fx111*=(double)idx; gx111*=(double)idx; ex111*=(double)idx;
                }
                
                if (idy!=1.0)
                {
                    fy000*=(double)idy; gy000*=(double)idy; ey000*=(double)idy;
                    fy001*=(double)idy; gy001*=(double)idy; ey001*=(double)idy;
                    fy010*=(double)idy; gy010*=(double)idy; ey010*=(double)idy;
                    fy011*=(double)idy; gy011*=(double)idy; ey011*=(double)idy;
                    fy100*=(double)idy; gy100*=(double)idy; ey100*=(double)idy;
                    fy101*=(double)idy; gy101*=(double)idy; ey101*=(double)idy;
                    fy110*=(double)idy; gy110*=(double)idy; ey110*=(double)idy;
                    fy111*=(double)idy; gy111*=(double)idy; ey111*=(double)idy;
                }
                
                if (idz!=1.0)
                {
                    fz000*=(double)idz; gz000*=(double)idz; ez000*=(double)idz;
                    fz001*=(double)idz; gz001*=(double)idz; ez001*=(double)idz;
                    fz010*=(double)idz; gz010*=(double)idz; ez010*=(double)idz;
                    fz011*=(double)idz; gz011*=(double)idz; ez011*=(double)idz;
                    fz100*=(double)idz; gz100*=(double)idz; ez100*=(double)idz;
                    fz101*=(double)idz; gz101*=(double)idz; ez101*=(double)idz;
                    fz110*=(double)idz; gz110*=(double)idz; ez110*=(double)idz;
                    fz111*=(double)idz; gz111*=(double)idz; ez111*=(double)idz;
                }
                
                if (disp)
                {
                    fx000+=1.0;
                    fx001+=1.0;
                    fx010+=1.0;
                    fx011+=1.0;
                    fx100+=1.0;
                    fx101+=1.0;
                    fx110+=1.0;
                    fx111+=1.0;
                    gy000+=1.0;
                    gy001+=1.0;
                    gy010+=1.0;
                    gy011+=1.0;
                    gy100+=1.0;
                    gy101+=1.0;
                    gy110+=1.0;
                    gy111+=1.0;
                    ez000+=1.0;
                    ez001+=1.0;
                    ez010+=1.0;
                    ez011+=1.0;
                    ez100+=1.0;
                    ez101+=1.0;
                    ez110+=1.0;
                    ez111+=1.0;
                }
                
                J000=fx000*(gy000*ez000-gz000*ey000)-fy000*(gx000*ez000-gz000*ex000)+fz000*(gx000*ey000-gy000*ex000);
                J001=fx001*(gy001*ez001-gz001*ey001)-fy001*(gx001*ez001-gz001*ex001)+fz001*(gx001*ey001-gy001*ex001);
                J010=fx010*(gy010*ez010-gz010*ey010)-fy010*(gx010*ez010-gz010*ex010)+fz010*(gx010*ey010-gy010*ex010);
                J011=fx011*(gy011*ez011-gz011*ey011)-fy011*(gx011*ez011-gz011*ex011)+fz011*(gx011*ey011-gy011*ex011);
                J100=fx100*(gy100*ez100-gz100*ey100)-fy100*(gx100*ez100-gz100*ex100)+fz100*(gx100*ey100-gy100*ex100);
                J101=fx101*(gy101*ez101-gz101*ey101)-fy101*(gx101*ez101-gz101*ex101)+fz101*(gx101*ey101-gy101*ex101);
                J110=fx110*(gy110*ez110-gz110*ey110)-fy110*(gx110*ez110-gz110*ex110)+fz110*(gx110*ey110-gy110*ex110);
                J111=fx111*(gy111*ez111-gz111*ey111)-fy111*(gx111*ez111-gz111*ex111)+fz111*(gx111*ey111-gy111*ex111);
                
                (*J)[i000]+=J000;
                (*J)[i001]+=J001;
                (*J)[i010]+=J010;
                (*J)[i011]+=J011;
                (*J)[i100]+=J100;
                (*J)[i101]+=J101;
                (*J)[i110]+=J110;
                (*J)[i111]+=J111;
                
                NB[i000]++;
                NB[i001]++;
                NB[i010]++;
                NB[i011]++;
                NB[i100]++;
                NB[i101]++;
                NB[i110]++;
                NB[i111]++;
            }
    for (i=0;i<rcs;i++)
        (*J)[i]/=((double)NB[i]);
    return 1;
}

/********************************************************************************************************************************/	
void Usage(char *programName)
{
    int i,n=0;
    char *emptyspace;
    
    while (programName[n]!='\0')
        n++;
    emptyspace=(char *)malloc(sizeof(char)*n);
    for (i=0;i<n;i++)
        emptyspace[i]=' ';
    emptyspace[n]='\0';
	fprintf(stderr,"%s usage:\n",programName);
	fprintf(stderr,"%s computes the field of Jacobian determinants pertaining to a given deformation. ",programName);
    fprintf(stderr,"Specifically, it returns the average of the corner Jacobians that exist at a voxel, for all voxels. \n");
    fprintf(stderr,"%s takes the following options:\n",programName);
    fprintf(stderr,"%s -d <string, deformation field> \n",emptyspace);
    fprintf(stderr,"%s -r <int, number of rows (y dim) [256]> \n",emptyspace);
    fprintf(stderr,"%s -c <int, number of columns (x dim) [256]> \n",emptyspace);
    fprintf(stderr,"%s -s <int, number of slices (z dim) [198]> \n",emptyspace);
    fprintf(stderr,"%s -Sx <float, voxel size along the x dimension [1.0]> \n",emptyspace);
    fprintf(stderr,"%s -Sy <float, voxel size along the y dimension [1.0]> \n",emptyspace);
    fprintf(stderr,"%s -Sz <float, voxel size along the z dimension [1.0]> \n",emptyspace);
    fprintf(stderr,"%s -D <0-1, 1 if the deformation encodes the displacements [1]> \n",emptyspace);
    fprintf(stderr,"%s -8 <for double precision output> \n",emptyspace);
    fprintf(stderr,"%s -o <string, output Jacobian field [defaultoutput.dat]> \n",emptyspace);
    
    free(emptyspace);
}

/********************************************************************************************************************************/	

/* returns the index of the first argument that is not an option; i.e.
   does not start with a dash or a slash
*/
int HandleOptions(int argc,char *argv[])
{
	int i,firstnonoption=0;
	float tau,popt;

	for (i=1; i< argc;i++) {
		if (argv[i][0] == '/' || argv[i][0] == '-') {
			switch (argv[i][1]) {
				/* An argument -? means help is requested */
				case '?':
					Usage(argv[0]);
					break;
/*				case 'h':
				case 'H':
					if (!stricmp(argv[i]+1,"help")) {
						Usage(argv[0]);
						break;
					}
					/* If the option -h means anything else
					 * in your application add code here
					 * Note: this falls through to the default
					 * to print an "unknow option" message
					*/
				/* add your option switches here */
				case 'o':
					strcpy((char *)&Outputfilename,argv[i+1]);
                    NOOUTFILE=0;
                    break;
				case 'd':
					strcpy((char *)&Deformationfilename,argv[i+1]);
                    NODEFORMATION=0;
                    break;
				case 'S':
                    if (argv[i][2]=='x')
                        iDX=1.0/atof(argv[i+1]);
                    if (argv[i][2]=='y')
                        iDY=1.0/atof(argv[i+1]);
                    if (argv[i][2]=='z')
                        iDZ=1.0/atof(argv[i+1]);
                    break;
				case 'r':
					R=atoi(argv[i+1]);
                    if (R<2)
                    {
                        fprintf(stderr,"number of rows too small (%d).\n",R);
                        fprintf(stderr,"bailing out ..\n");
                        return 0;
                    }
					break;
				case 'c':
    			    C=atoi(argv[i+1]);
                    if (C<2)
                    {
                        fprintf(stderr,"number of columns too small (%d).\n",C);
                        fprintf(stderr,"bailing out ..\n");
                        return 0;
                    }
					break;
				case 's':
	    			S=atoi(argv[i+1]);
                    if (S<2)
                    {
                        fprintf(stderr,"number of slices too small (%d).\n",S);
                        fprintf(stderr,"bailing out ..\n");
                        return 0;
                    }
					break;
				case 'D':
	    			DISP=atoi(argv[i+1]);
                    if ((DISP<0)||(DISP>1))
                    {
                        fprintf(stderr,"illegal value for -D (%d).\n",DISP);
                        fprintf(stderr,"bailing out ..\n");
                        return 0;
                    }
					break;
				case '8':
	    			DOUBLEOUTPUT=1;
					break;
				default:
					fprintf(stderr,"unknown option %s\n",argv[i]);
					break;
			}
		}
	}
	return 1;
}



/********************************************************************************************************************************/	

int main(int argc,char *argv[])
{
    float *f, *g, *e;
    double *J, dtemp;
	int i,j,rcs;
	float ftemp;
    FILE *Imagefile, *Outfile;
	
	if (argc == 1) {
		/* If no arguments we call the Usage routine and exit */
		Usage(argv[0]);
		return 1;
	}
	/* handle the program options */
	if (HandleOptions(argc,argv))
    {
        rcs=R*C*S;
        if (NODEFORMATION)
        {
            fprintf(stderr,"problem with the input file\n");
            fprintf(stderr,"bailing out ..\n");
            return 0;
        } else
            if (LoadDeformationFieldCartesianComponents3D(Deformationfilename,&f,&g,&e,R,C,S))
            {
                /* check output file */
                if (NOOUTFILE)
                {
                    if ((Outfile=fopen(Defaultoutputfilename,"wb"))==NULL)
                    {
                        fprintf(stderr,"Cannot open file %s for write.\n",Outputfilename);
                        return 0;
                    } else
                        fclose(Outfile);
                } else
                {
                    if ((Outfile=fopen(Outputfilename,"wb"))==NULL)
                    {
                        fprintf(stderr,"Cannot open file %s for write.\n",Outputfilename);
                        return 0;
                    } else
                        fclose(Outfile);
                }
 
                J=(double *)calloc(rcs,sizeof(double));
 
                DeformationJacobian3D(f,g,e,&J,R,C,S,DISP,iDX,iDY,iDZ);
 
                /* saving the result */
            	if (NOOUTFILE)
                    if ((Outfile=fopen(Defaultoutputfilename,"wb"))==NULL)
                    {
                        fprintf(stderr,"cannot open the image file %s for write\n",Defaultoutputfilename);
                        fprintf(stderr,"warped image not written.\n");
                    } else
                    {
                        if (DOUBLEOUTPUT)
                        {
                            for (i=0;i<rcs;i++)
                            {
                                dtemp=J[i];
                                fwrite(&dtemp,sizeof(double),1,Outfile);
                            }
                            fclose(Outfile);
                        } else
                        {
                            for (i=0;i<rcs;i++)
                            {
                                ftemp=(float)J[i];
                                fwrite(&ftemp,sizeof(float),1,Outfile);
                            }
                            fclose(Outfile);
                        }
                    }
    	        else
                    if ((Outfile=fopen(Outputfilename,"wb"))==NULL)
                    {
                        fprintf(stderr,"cannot open the image file %s for write\n",Outputfilename);
                        fprintf(stderr,"warped image not written.\n");
                    } else
                    {
                        if (DOUBLEOUTPUT)
                        {
                            for (i=0;i<rcs;i++)
                            {
                                dtemp=J[i];
                                fwrite(&dtemp,sizeof(double),1,Outfile);
                            }
                            fclose(Outfile);
                        } else
                        {
                            for (i=0;i<rcs;i++)
                            {
                                ftemp=(float)J[i];
                                fwrite(&ftemp,sizeof(float),1,Outfile);
                            }
                            fclose(Outfile);
                        }
                    }
 
 
            	/* freeing the emmory allocated to f, g, and e */
                free(f);
                free(g);
                free(e);
                free(J);
                return 1;
            } else
            {
                fprintf(stderr,"no action performed.\n");
        	    return 0;
            }
    } else
    {
        fprintf(stderr,"no action performed.\n");
    	return 0;
    }
}

