/********************************************************************************************************************************/      
/* Disclaimer
/* This code is provided as is, without any guarantees. Published manuscripts using this code should cite
/* the following article:
/*      B. Kara{\c{c}}al{\i}, C. Davatzikos, ``{\em Simulation of tissue atrophy using a topology preserving 
/*        transformation model}'' 
/* Comments and bug reports should be forwarded to bilge@rad.upenn.edu.
/* VoxelVolumeMatching3D -- Copyright Bilge Karacali 2005
/********************************************************************************************************************************/      

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <dfieldio.h>

/********************************************************************************************************************************/	

/* Predefined constants */
#define PI 3.14159265358979

/* macros */
#define JacobianPenalty(x,y) ((x)/(y)+(y)/(x)-2.0)

/********************************************************************************************************************************/	

/* Global Variables */
int r=256;
int c=256;
int s=198;
int BS=4;
char Deformationfieldfilename[600];
char Volumefilename[600];
char Outputfilename[600];
char *Defaultoutputfilename="defaultoutput.dat";
char Restrictionfilename[600];
int NOOUTFILE=1;
float DX=1.0;
float DY=1.0;
float DZ=1.0;
int MAXITNUM=1000;
int NORESTRICTION=1;
int VERBOSITY=1;
int NODEFORMATION=1;
int DF=1;
float LAMBDA=1.0;
float JMIN=0.3;

/********************************************************************************************************************************/	

/* Function Declarations */
void Usage();
int HandleOptions();
float VoxelVolumeMatchingCost3D();
int VoxelVolumeMatchingGrads3D();
int VoxelVolumeMatching3D(float **, float **, float **, float *, int, int, int, int, int, unsigned char *, float, float);
float TopologyCost3D(float *, float *, float *, int , int , int , float , float );
int TopologyGrads3D(float *, float *, float *, int , int , int , float , float , float **, float **, float **);

/********************************************************************************************************************************/	
/* Function Definitions */

float TopologyCost3D(float *f, float *g, float *e, int r, int c, int s, float Jmin, float lambda)
{
    float res=0.0;
    int i, j, k, rc=r*c;
    int i000, i001, i010, i011, i100, i101, i110, i111;
    float f000, f001, f010, f011, f100, f101, f110, f111;
    float fx000, fx001, fx010, fx011, fx100, fx101, fx110, fx111;
    float fy000, fy001, fy010, fy011, fy100, fy101, fy110, fy111;
    float fz000, fz001, fz010, fz011, fz100, fz101, fz110, fz111;
    float g000, g001, g010, g011, g100, g101, g110, g111;
    float gx000, gx001, gx010, gx011, gx100, gx101, gx110, gx111;
    float gy000, gy001, gy010, gy011, gy100, gy101, gy110, gy111;
    float gz000, gz001, gz010, gz011, gz100, gz101, gz110, gz111;
    float e000, e001, e010, e011, e100, e101, e110, e111;
    float ex000, ex001, ex010, ex011, ex100, ex101, ex110, ex111;
    float ey000, ey001, ey010, ey011, ey100, ey101, ey110, ey111;
    float ez000, ez001, ez010, ez011, ez100, ez101, ez110, ez111;
    float J000, J001, J010, J011, J100, J101, J110, J111;
    
    for (k=0;k<s-1;k++)
        for (j=0;j<c-1;j++)
            for (i=0;i<r-1;i++)
            {
                i000=k*rc+j*r+i;
                i001=i000+r;
                i010=i000+1;
                i011=i001+1;
                i100=i000+rc;
                i101=i100+r;
                i110=i100+1;
                i111=i101+1;
                
                f000=f[i000]; f001=f[i001]; f010=f[i010]; f011=f[i011]; f100=f[i100]; f101=f[i101]; f110=f[i110]; f111=f[i111]; 
                g000=g[i000]; g001=g[i001]; g010=g[i010]; g011=g[i011]; g100=g[i100]; g101=g[i101]; g110=g[i110]; g111=g[i111]; 
                e000=e[i000]; e001=e[i001]; e010=e[i010]; e011=e[i011]; e100=e[i100]; e101=e[i101]; e110=e[i110]; e111=e[i111]; 
                
                fx000=f001-f000; fy000=f010-f000; fz000=f100-f000; fx001=f001-f000; fy001=f011-f001; fz001=f101-f001; fx010=f011-f010; fy010=f010-f000; fz010=f110-f010; fx011=f011-f010; fy011=f011-f001; fz011=f111-f011; fx100=f101-f100; fy100=f110-f100; fz100=f100-f000; fx101=f101-f100; fy101=f111-f101; fz101=f101-f001; fx110=f111-f110; fy110=f110-f100; fz110=f110-f010; fx111=f111-f110; fy111=f111-f101; fz111=f111-f011; 
                gx000=g001-g000; gy000=g010-g000; gz000=g100-g000; gx001=g001-g000; gy001=g011-g001; gz001=g101-g001; gx010=g011-g010; gy010=g010-g000; gz010=g110-g010; gx011=g011-g010; gy011=g011-g001; gz011=g111-g011; gx100=g101-g100; gy100=g110-g100; gz100=g100-g000; gx101=g101-g100; gy101=g111-g101; gz101=g101-g001; gx110=g111-g110; gy110=g110-g100; gz110=g110-g010; gx111=g111-g110; gy111=g111-g101; gz111=g111-g011; 
                ex000=e001-e000; ey000=e010-e000; ez000=e100-e000; ex001=e001-e000; ey001=e011-e001; ez001=e101-e001; ex010=e011-e010; ey010=e010-e000; ez010=e110-e010; ex011=e011-e010; ey011=e011-e001; ez011=e111-e011; ex100=e101-e100; ey100=e110-e100; ez100=e100-e000; ex101=e101-e100; ey101=e111-e101; ez101=e101-e001; ex110=e111-e110; ey110=e110-e100; ez110=e110-e010; ex111=e111-e110; ey111=e111-e101; ez111=e111-e011; 
                
                J000=fx000*(gy000*ez000-gz000*ey000)-fy000*(gx000*ez000-gz000*ex000)+fz000*(gx000*ey000-gy000*ex000);
                J001=fx001*(gy001*ez001-gz001*ey001)-fy001*(gx001*ez001-gz001*ex001)+fz001*(gx001*ey001-gy001*ex001);
                J010=fx010*(gy010*ez010-gz010*ey010)-fy010*(gx010*ez010-gz010*ex010)+fz010*(gx010*ey010-gy010*ex010);
                J011=fx011*(gy011*ez011-gz011*ey011)-fy011*(gx011*ez011-gz011*ex011)+fz011*(gx011*ey011-gy011*ex011);
                J100=fx100*(gy100*ez100-gz100*ey100)-fy100*(gx100*ez100-gz100*ex100)+fz100*(gx100*ey100-gy100*ex100);
                J101=fx101*(gy101*ez101-gz101*ey101)-fy101*(gx101*ez101-gz101*ex101)+fz101*(gx101*ey101-gy101*ex101);
                J110=fx110*(gy110*ez110-gz110*ey110)-fy110*(gx110*ez110-gz110*ex110)+fz110*(gx110*ey110-gy110*ex110);
                J111=fx111*(gy111*ez111-gz111*ey111)-fy111*(gx111*ez111-gz111*ex111)+fz111*(gx111*ey111-gy111*ex111);
                
                if (J000<Jmin) res+=JacobianPenalty(J000,Jmin);
                if (J001<Jmin) res+=JacobianPenalty(J001,Jmin);
                if (J010<Jmin) res+=JacobianPenalty(J010,Jmin);
                if (J011<Jmin) res+=JacobianPenalty(J011,Jmin);
                if (J100<Jmin) res+=JacobianPenalty(J100,Jmin);
                if (J101<Jmin) res+=JacobianPenalty(J101,Jmin);
                if (J110<Jmin) res+=JacobianPenalty(J110,Jmin);
                if (J111<Jmin) res+=JacobianPenalty(J111,Jmin);
            }
    res*=lambda;
    
    return res;
}

/********************************************************************************************************************************/	

int TopologyGrads3D(float *f, float *g, float *e, int r, int c, int s, float Jmin, float lambda, float **df, float **dg, float **de)
{
    float res=0.0;
    int i, j, k, rc=r*c;
    int i000, i001, i010, i011, i100, i101, i110, i111;
    float f000, f001, f010, f011, f100, f101, f110, f111;
    float fx000, fx001, fx010, fx011, fx100, fx101, fx110, fx111;
    float fy000, fy001, fy010, fy011, fy100, fy101, fy110, fy111;
    float fz000, fz001, fz010, fz011, fz100, fz101, fz110, fz111;
    float g000, g001, g010, g011, g100, g101, g110, g111;
    float gx000, gx001, gx010, gx011, gx100, gx101, gx110, gx111;
    float gy000, gy001, gy010, gy011, gy100, gy101, gy110, gy111;
    float gz000, gz001, gz010, gz011, gz100, gz101, gz110, gz111;
    float e000, e001, e010, e011, e100, e101, e110, e111;
    float ex000, ex001, ex010, ex011, ex100, ex101, ex110, ex111;
    float ey000, ey001, ey010, ey011, ey100, ey101, ey110, ey111;
    float ez000, ez001, ez010, ez011, ez100, ez101, ez110, ez111;
    float J000, J001, J010, J011, J100, J101, J110, J111;
    
    for (k=0;k<s-1;k++)
        for (j=0;j<c-1;j++)
            for (i=0;i<r-1;i++)
            {
                i000=k*rc+j*r+i;
                i001=i000+r;
                i010=i000+1;
                i011=i001+1;
                i100=i000+rc;
                i101=i100+r;
                i110=i100+1;
                i111=i101+1;
                
                f000=f[i000]; f001=f[i001]; f010=f[i010]; f011=f[i011]; f100=f[i100]; f101=f[i101]; f110=f[i110]; f111=f[i111]; 
                g000=g[i000]; g001=g[i001]; g010=g[i010]; g011=g[i011]; g100=g[i100]; g101=g[i101]; g110=g[i110]; g111=g[i111]; 
                e000=e[i000]; e001=e[i001]; e010=e[i010]; e011=e[i011]; e100=e[i100]; e101=e[i101]; e110=e[i110]; e111=e[i111]; 
                
                fx000=f001-f000; fy000=f010-f000; fz000=f100-f000; fx001=f001-f000; fy001=f011-f001; fz001=f101-f001; fx010=f011-f010; fy010=f010-f000; fz010=f110-f010; fx011=f011-f010; fy011=f011-f001; fz011=f111-f011; fx100=f101-f100; fy100=f110-f100; fz100=f100-f000; fx101=f101-f100; fy101=f111-f101; fz101=f101-f001; fx110=f111-f110; fy110=f110-f100; fz110=f110-f010; fx111=f111-f110; fy111=f111-f101; fz111=f111-f011; 
                gx000=g001-g000; gy000=g010-g000; gz000=g100-g000; gx001=g001-g000; gy001=g011-g001; gz001=g101-g001; gx010=g011-g010; gy010=g010-g000; gz010=g110-g010; gx011=g011-g010; gy011=g011-g001; gz011=g111-g011; gx100=g101-g100; gy100=g110-g100; gz100=g100-g000; gx101=g101-g100; gy101=g111-g101; gz101=g101-g001; gx110=g111-g110; gy110=g110-g100; gz110=g110-g010; gx111=g111-g110; gy111=g111-g101; gz111=g111-g011; 
                ex000=e001-e000; ey000=e010-e000; ez000=e100-e000; ex001=e001-e000; ey001=e011-e001; ez001=e101-e001; ex010=e011-e010; ey010=e010-e000; ez010=e110-e010; ex011=e011-e010; ey011=e011-e001; ez011=e111-e011; ex100=e101-e100; ey100=e110-e100; ez100=e100-e000; ex101=e101-e100; ey101=e111-e101; ez101=e101-e001; ex110=e111-e110; ey110=e110-e100; ez110=e110-e010; ex111=e111-e110; ey111=e111-e101; ez111=e111-e011; 
                
                J000=fx000*(gy000*ez000-gz000*ey000)-fy000*(gx000*ez000-gz000*ex000)+fz000*(gx000*ey000-gy000*ex000);
                J001=fx001*(gy001*ez001-gz001*ey001)-fy001*(gx001*ez001-gz001*ex001)+fz001*(gx001*ey001-gy001*ex001);
                J010=fx010*(gy010*ez010-gz010*ey010)-fy010*(gx010*ez010-gz010*ex010)+fz010*(gx010*ey010-gy010*ex010);
                J011=fx011*(gy011*ez011-gz011*ey011)-fy011*(gx011*ez011-gz011*ex011)+fz011*(gx011*ey011-gy011*ex011);
                J100=fx100*(gy100*ez100-gz100*ey100)-fy100*(gx100*ez100-gz100*ex100)+fz100*(gx100*ey100-gy100*ex100);
                J101=fx101*(gy101*ez101-gz101*ey101)-fy101*(gx101*ez101-gz101*ex101)+fz101*(gx101*ey101-gy101*ex101);
                J110=fx110*(gy110*ez110-gz110*ey110)-fy110*(gx110*ez110-gz110*ex110)+fz110*(gx110*ey110-gy110*ex110);
                J111=fx111*(gy111*ez111-gz111*ey111)-fy111*(gx111*ez111-gz111*ex111)+fz111*(gx111*ey111-gy111*ex111);
                
                if (J000<Jmin)
                {
                    (*df)[i000] += lambda*(-Jmin/(J000*J000)+1/Jmin)*(-g010*e100+g100*e010+g001*e100-g100*e001-g001*e010+g010*e001);
                    (*df)[i001] += lambda*(-Jmin/(J000*J000)+1/Jmin)*((g010-g000)*(e100-e000)-(g100-g000)*(e010-e000));
                    (*df)[i010] += lambda*(-Jmin/(J000*J000)+1/Jmin)*(-(g001-g000)*(e100-e000)+(g100-g000)*(e001-e000));
                    (*df)[i100] += lambda*(-Jmin/(J000*J000)+1/Jmin)*((g001-g000)*(e010-e000)-(g010-g000)*(e001-e000));
                    (*dg)[i000] += lambda*(-Jmin/(J000*J000)+1/Jmin)*(f001*(-e100+e010)-f010*(-e100+e001)+f100*(-e010+e001));
                    (*dg)[i001] += lambda*(-Jmin/(J000*J000)+1/Jmin)*(-(f010-f000)*(e100-e000)+(f100-f000)*(e010-e000));
                    (*dg)[i010] += lambda*(-Jmin/(J000*J000)+1/Jmin)*((f001-f000)*(e100-e000)+(f100-f000)*(-e001+e000));
                    (*dg)[i100] += lambda*(-Jmin/(J000*J000)+1/Jmin)*((f001-f000)*(-e010+e000)-(f010-f000)*(-e001+e000));
                    (*de)[i000] += lambda*(-Jmin/(J000*J000)+1/Jmin)*(-f001*g010+f001*g100+f010*g001-f010*g100-f100*g001+f100*g010);
                    (*de)[i001] += lambda*(-Jmin/(J000*J000)+1/Jmin)*(-(f010-f000)*(-g100+g000)+(f100-f000)*(-g010+g000));
                    (*de)[i010] += lambda*(-Jmin/(J000*J000)+1/Jmin)*((f001-f000)*(-g100+g000)+(f100-f000)*(g001-g000));
                    (*de)[i100] += lambda*(-Jmin/(J000*J000)+1/Jmin)*((f001-f000)*(g010-g000)-(f010-f000)*(g001-g000));
                }
                if (J001<Jmin)
                {
                    (*df)[i000] += lambda*(-Jmin/(J001*J001)+1/Jmin)*(-(g011-g001)*(e101-e001)+(g101-g001)*(e011-e001));
                    (*df)[i001] += lambda*(-Jmin/(J001*J001)+1/Jmin)*(g011*e101-g101*e011-g000*e101+g101*e000+g000*e011-g011*e000);
                    (*df)[i011] += lambda*(-Jmin/(J001*J001)+1/Jmin)*(-(g001-g000)*(e101-e001)+(g101-g001)*(e001-e000));
                    (*df)[i101] += lambda*(-Jmin/(J001*J001)+1/Jmin)*((g001-g000)*(e011-e001)-(g011-g001)*(e001-e000));
                    (*dg)[i000] += lambda*(-Jmin/(J001*J001)+1/Jmin)*(-(f011-f001)*(-e101+e001)+(f101-f001)*(-e011+e001));
                    (*dg)[i001] += lambda*(-Jmin/(J001*J001)+1/Jmin)*(f000*e101-f000*e011-f011*e101+f011*e000+f101*e011-f101*e000);
                    (*dg)[i011] += lambda*(-Jmin/(J001*J001)+1/Jmin)*((f001-f000)*(e101-e001)+(f101-f001)*(-e001+e000));
                    (*dg)[i101] += lambda*(-Jmin/(J001*J001)+1/Jmin)*((f001-f000)*(-e011+e001)-(f011-f001)*(-e001+e000));
                    (*de)[i000] += lambda*(-Jmin/(J001*J001)+1/Jmin)*(-(f011-f001)*(g101-g001)+(f101-f001)*(g011-g001));
                    (*de)[i001] += lambda*(-Jmin/(J001*J001)+1/Jmin)*(-f000*g101+f000*g011-f011*g000+f011*g101+f101*g000-f101*g011);
                    (*de)[i011] += lambda*(-Jmin/(J001*J001)+1/Jmin)*((f001-f000)*(-g101+g001)+(f101-f001)*(g001-g000));
                    (*de)[i101] += lambda*(-Jmin/(J001*J001)+1/Jmin)*((f001-f000)*(g011-g001)-(f011-f001)*(g001-g000));
                }
                if (J010<Jmin)
                {
                    (*df)[i000] += lambda*(-Jmin/(J010*J010)+1/Jmin)*((g011-g010)*(e110-e010)-(g110-g010)*(e011-e010));
                    (*df)[i010] += lambda*(-Jmin/(J010*J010)+1/Jmin)*(g000*e110-g110*e000-g011*e110+g110*e011+g011*e000-g000*e011);
                    (*df)[i011] += lambda*(-Jmin/(J010*J010)+1/Jmin)*((g010-g000)*(e110-e010)-(g110-g010)*(e010-e000));
                    (*df)[i110] += lambda*(-Jmin/(J010*J010)+1/Jmin)*((g011-g010)*(e010-e000)-(g010-g000)*(e011-e010));
                    (*dg)[i000] += lambda*(-Jmin/(J010*J010)+1/Jmin)*((f011-f010)*(-e110+e010)+(f110-f010)*(e011-e010));
                    (*dg)[i010] += lambda*(-Jmin/(J010*J010)+1/Jmin)*(f011*e110-f011*e000-f000*e110+f000*e011-f110*e011+f110*e000);
                    (*dg)[i011] += lambda*(-Jmin/(J010*J010)+1/Jmin)*(-(f010-f000)*(e110-e010)+(f110-f010)*(e010-e000));
                    (*dg)[i110] += lambda*(-Jmin/(J010*J010)+1/Jmin)*((f011-f010)*(-e010+e000)-(f010-f000)*(-e011+e010));
                    (*de)[i000] += lambda*(-Jmin/(J010*J010)+1/Jmin)*((f011-f010)*(g110-g010)+(f110-f010)*(-g011+g010));
                    (*de)[i010] += lambda*(-Jmin/(J010*J010)+1/Jmin)*(f011*g000-f011*g110+f000*g110-f000*g011-f110*g000+f110*g011);
                    (*de)[i011] += lambda*(-Jmin/(J010*J010)+1/Jmin)*(-(f010-f000)*(-g110+g010)+(f110-f010)*(-g010+g000));
                    (*de)[i110] += lambda*(-Jmin/(J010*J010)+1/Jmin)*((f011-f010)*(g010-g000)-(f010-f000)*(g011-g010));
                }
                if (J011<Jmin)
                {
                    (*df)[i001] += lambda*(-Jmin/(J011*J011)+1/Jmin)*((g011-g010)*(e111-e011)-(g111-g011)*(e011-e010));
                    (*df)[i010] += lambda*(-Jmin/(J011*J011)+1/Jmin)*(-(g011-g001)*(e111-e011)+(g111-g011)*(e011-e001));
                    (*df)[i011] += lambda*(-Jmin/(J011*J011)+1/Jmin)*(-g001*e111+g111*e001+g010*e111-g111*e010-g010*e001+g001*e010);
                    (*df)[i111] += lambda*(-Jmin/(J011*J011)+1/Jmin)*((g011-g010)*(e011-e001)-(g011-g001)*(e011-e010));
                    (*dg)[i001] += lambda*(-Jmin/(J011*J011)+1/Jmin)*((f011-f010)*(-e111+e011)+(f111-f011)*(e011-e010));
                    (*dg)[i010] += lambda*(-Jmin/(J011*J011)+1/Jmin)*(-(f011-f001)*(-e111+e011)+(f111-f011)*(-e011+e001));
                    (*dg)[i011] += lambda*(-Jmin/(J011*J011)+1/Jmin)*(-f010*e111+f010*e001+f001*e111-f001*e010-f111*e001+f111*e010);
                    (*dg)[i111] += lambda*(-Jmin/(J011*J011)+1/Jmin)*((f011-f010)*(-e011+e001)-(f011-f001)*(-e011+e010));
                    (*de)[i001] += lambda*(-Jmin/(J011*J011)+1/Jmin)*((f011-f010)*(g111-g011)+(f111-f011)*(-g011+g010));
                    (*de)[i010] += lambda*(-Jmin/(J011*J011)+1/Jmin)*(-(f011-f001)*(g111-g011)+(f111-f011)*(g011-g001));
                    (*de)[i011] += lambda*(-Jmin/(J011*J011)+1/Jmin)*(-f010*g001+f010*g111+f001*g010-f001*g111+f111*g001-f111*g010);
                    (*de)[i111] += lambda*(-Jmin/(J011*J011)+1/Jmin)*((f011-f010)*(g011-g001)-(f011-f001)*(g011-g010));
                }
                if (J100<Jmin)
                {
                    (*df)[i000] += lambda*(-Jmin/(J100*J100)+1/Jmin)*(-(g101-g100)*(e110-e100)+(g110-g100)*(e101-e100));
                    (*df)[i100] += lambda*(-Jmin/(J100*J100)+1/Jmin)*(g110*e000-g000*e110-g101*e000+g000*e101+g101*e110-g110*e101);
                    (*df)[i101] += lambda*(-Jmin/(J100*J100)+1/Jmin)*((g110-g100)*(e100-e000)-(g100-g000)*(e110-e100));
                    (*df)[i110] += lambda*(-Jmin/(J100*J100)+1/Jmin)*(-(g101-g100)*(e100-e000)+(g100-g000)*(e101-e100));
                    (*dg)[i000] += lambda*(-Jmin/(J100*J100)+1/Jmin)*((f101-f100)*(e110-e100)-(f110-f100)*(e101-e100));
                    (*dg)[i100] += lambda*(-Jmin/(J100*J100)+1/Jmin)*(f101*e000-f101*e110-f110*e000+f110*e101-f000*e101+f000*e110);
                    (*dg)[i101] += lambda*(-Jmin/(J100*J100)+1/Jmin)*(-(f110-f100)*(e100-e000)+(f100-f000)*(e110-e100));
                    (*dg)[i110] += lambda*(-Jmin/(J100*J100)+1/Jmin)*((f101-f100)*(e100-e000)+(f100-f000)*(-e101+e100));
                    (*de)[i000] += lambda*(-Jmin/(J100*J100)+1/Jmin)*((f101-f100)*(-g110+g100)-(f110-f100)*(-g101+g100));
                    (*de)[i100] += lambda*(-Jmin/(J100*J100)+1/Jmin)*(f101*g110-f101*g000-f110*g101+f110*g000+f000*g101-f000*g110);
                    (*de)[i101] += lambda*(-Jmin/(J100*J100)+1/Jmin)*(-(f110-f100)*(-g100+g000)+(f100-f000)*(-g110+g100));
                    (*de)[i110] += lambda*(-Jmin/(J100*J100)+1/Jmin)*((f101-f100)*(-g100+g000)+(f100-f000)*(g101-g100));
                }
                if (J101<Jmin)
                {
                    (*df)[i001] += lambda*(-Jmin/(J101*J101)+1/Jmin)*(-(g101-g100)*(e111-e101)+(g111-g101)*(e101-e100));
                    (*df)[i100] += lambda*(-Jmin/(J101*J101)+1/Jmin)*(-(g111-g101)*(e101-e001)+(g101-g001)*(e111-e101));
                    (*df)[i101] += lambda*(-Jmin/(J101*J101)+1/Jmin)*(-g111*e001+g001*e111+g100*e001-g001*e100-g100*e111+g111*e100);
                    (*df)[i111] += lambda*(-Jmin/(J101*J101)+1/Jmin)*(-(g101-g100)*(e101-e001)+(g101-g001)*(e101-e100));
                    (*dg)[i001] += lambda*(-Jmin/(J101*J101)+1/Jmin)*((f101-f100)*(e111-e101)-(f111-f101)*(e101-e100));
                    (*dg)[i100] += lambda*(-Jmin/(J101*J101)+1/Jmin)*(-(f111-f101)*(-e101+e001)+(f101-f001)*(-e111+e101));
                    (*dg)[i101] += lambda*(-Jmin/(J101*J101)+1/Jmin)*(-f100*e001+f100*e111+f111*e001-f111*e100+f001*e100-f001*e111);
                    (*dg)[i111] += lambda*(-Jmin/(J101*J101)+1/Jmin)*((f101-f100)*(e101-e001)+(f101-f001)*(-e101+e100));
                    (*de)[i001] += lambda*(-Jmin/(J101*J101)+1/Jmin)*((f101-f100)*(-g111+g101)-(f111-f101)*(-g101+g100));
                    (*de)[i100] += lambda*(-Jmin/(J101*J101)+1/Jmin)*(-(f111-f101)*(g101-g001)+(f101-f001)*(g111-g101));
                    (*de)[i101] += lambda*(-Jmin/(J101*J101)+1/Jmin)*(-f100*g111+f100*g001+f111*g100-f111*g001-f001*g100+f001*g111);
                    (*de)[i111] += lambda*(-Jmin/(J101*J101)+1/Jmin)*((f101-f100)*(-g101+g001)+(f101-f001)*(g101-g100));
                }
                if (J110<Jmin)
                {
                    (*df)[i010] += lambda*(-Jmin/(J110*J110)+1/Jmin)*(-(g111-g110)*(e110-e100)+(g110-g100)*(e111-e110));
                    (*df)[i100] += lambda*(-Jmin/(J110*J110)+1/Jmin)*((g111-g110)*(e110-e010)-(g110-g010)*(e111-e110));
                    (*df)[i110] += lambda*(-Jmin/(J110*J110)+1/Jmin)*(-g100*e010+g010*e100+g111*e010-g010*e111-g111*e100+g100*e111);
                    (*df)[i111] += lambda*(-Jmin/(J110*J110)+1/Jmin)*((g110-g100)*(e110-e010)-(g110-g010)*(e110-e100));
                    (*dg)[i010] += lambda*(-Jmin/(J110*J110)+1/Jmin)*((f111-f110)*(e110-e100)-(f110-f100)*(e111-e110));
                    (*dg)[i100] += lambda*(-Jmin/(J110*J110)+1/Jmin)*((f111-f110)*(-e110+e010)+(f110-f010)*(e111-e110));
                    (*dg)[i110] += lambda*(-Jmin/(J110*J110)+1/Jmin)*(-f111*e010+f111*e100+f100*e010-f100*e111+f010*e111-f010*e100);
                    (*dg)[i111] += lambda*(-Jmin/(J110*J110)+1/Jmin)*(-(f110-f100)*(e110-e010)+(f110-f010)*(e110-e100));
                    (*de)[i010] += lambda*(-Jmin/(J110*J110)+1/Jmin)*((f111-f110)*(-g110+g100)-(f110-f100)*(-g111+g110));
                    (*de)[i100] += lambda*(-Jmin/(J110*J110)+1/Jmin)*((f111-f110)*(g110-g010)+(f110-f010)*(-g111+g110));
                    (*de)[i110] += lambda*(-Jmin/(J110*J110)+1/Jmin)*(-f111*g100+f111*g010+f100*g111-f100*g010+f010*g100-f010*g111);
                    (*de)[i111] += lambda*(-Jmin/(J110*J110)+1/Jmin)*(-(f110-f100)*(-g110+g010)+(f110-f010)*(-g110+g100));
                }
                if (J111<Jmin)
                {
                    (*df)[i011] += lambda*(-Jmin/(J111*J111)+1/Jmin)*(-(g111-g110)*(e111-e101)+(g111-g101)*(e111-e110));
                    (*df)[i101] += lambda*(-Jmin/(J111*J111)+1/Jmin)*((g111-g110)*(e111-e011)-(g111-g011)*(e111-e110));
                    (*df)[i110] += lambda*(-Jmin/(J111*J111)+1/Jmin)*(-(g111-g101)*(e111-e011)+(g111-g011)*(e111-e101));
                    (*df)[i111] += lambda*(-Jmin/(J111*J111)+1/Jmin)*(g101*e011-g011*e101-g110*e011+g011*e110+g110*e101-g101*e110);
                    (*dg)[i011] += lambda*(-Jmin/(J111*J111)+1/Jmin)*((f111-f110)*(e111-e101)-(f111-f101)*(e111-e110));
                    (*dg)[i101] += lambda*(-Jmin/(J111*J111)+1/Jmin)*((f111-f110)*(-e111+e011)+(f111-f011)*(e111-e110));
                    (*dg)[i110] += lambda*(-Jmin/(J111*J111)+1/Jmin)*(-(f111-f101)*(-e111+e011)+(f111-f011)*(-e111+e101));
                    (*dg)[i111] += lambda*(-Jmin/(J111*J111)+1/Jmin)*(f110*e011-f110*e101-f101*e011+f101*e110-f011*e110+f011*e101);
                    (*de)[i011] += lambda*(-Jmin/(J111*J111)+1/Jmin)*((f111-f110)*(-g111+g101)-(f111-f101)*(-g111+g110));
                    (*de)[i101] += lambda*(-Jmin/(J111*J111)+1/Jmin)*((f111-f110)*(g111-g011)+(f111-f011)*(-g111+g110));
                    (*de)[i110] += lambda*(-Jmin/(J111*J111)+1/Jmin)*(-(f111-f101)*(g111-g011)+(f111-f011)*(g111-g101));
                    (*de)[i111] += lambda*(-Jmin/(J111*J111)+1/Jmin)*(f110*g101-f110*g011-f101*g110+f101*g011+f011*g110-f011*g101);
                }
            }
    return i111;
}

/********************************************************************************************************************************/	
float VoxelVolumeMatchingCost3D(float *f, float *g, float *e, float *TV, int r, int c, int s)
{
    float res=0.0, cres;
    int i, j, k, rc=r*c, NB;
    int i000;
    float cf, cg, ce;
    float fx0, fx1, gx0, gx1, ex0, ex1;
    float fy0, fy1, gy0, gy1, ey0, ey1;
    float fz0, fz1, gz0, gz1, ez0, ez1;
    float J000, J001, J010, J011, J100, J101, J110, J111;
    float cV, cTV;
    int LOOP=1;
    
    for (k=0;(k<s)&&(LOOP);k++)
        for (j=0;(j<c)&&(LOOP);j++)
            for (i=0;(i<r)&&(LOOP);i++)
            {
                i000=k*rc+j*r+i;
                cTV=TV[i000];
                cf=f[i000]; cg=g[i000]; ce=e[i000]; 
                
                if (i>0)
                {
                    fy0=cf-f[i000-1];
                    gy0=cg-g[i000-1];
                    ey0=ce-e[i000-1];
                }
                if (i<r-1)
                {
                    fy1=f[i000+1]-cf;
                    gy1=g[i000+1]-cg;
                    ey1=e[i000+1]-ce;
                }
                if (j>0)
                {
                    fx0=cf-f[i000-r];
                    gx0=cg-g[i000-r];
                    ex0=ce-e[i000-r];
                }
                if (j<c-1)
                {
                    fx1=f[i000+r]-cf;
                    gx1=g[i000+r]-cg;
                    ex1=e[i000+r]-ce;
                }
                if (k>0)
                {
                    fz0=cf-f[i000-rc];
                    gz0=cg-g[i000-rc];
                    ez0=ce-e[i000-rc];
                }
                if (k<s-1)
                {
                    fz1=f[i000+rc]-cf;
                    gz1=g[i000+rc]-cg;
                    ez1=e[i000+rc]-ce;
                }
                
                cres=0.0;
                NB=0;
                cV=0.0;
                if (k>0)
                {
                    if (i>0)
                    {
                        if (j>0)
                        {
                            NB++;
                            J000=fx0*(gy0*ez0-gz0*ey0)-fy0*(gx0*ez0-gz0*ex0)+fz0*(gx0*ey0-gy0*ex0);
                            cV+=J000;
                            if (J000<0.0) LOOP=0;
                        }
                        if (j<c-1)
                        {
                            NB++;
                            J001=fx1*(gy0*ez0-gz0*ey0)-fy0*(gx1*ez0-gz0*ex1)+fz0*(gx1*ey0-gy0*ex1);
                            cV+=J001;
                            if (J001<0.0) LOOP=0;
                        }
                    }
                    if (i<r-1)
                    {
                        if (j>0)
                        {
                            NB++;
                            J010=fx0*(gy1*ez0-gz0*ey1)-fy1*(gx0*ez0-gz0*ex0)+fz0*(gx0*ey1-gy1*ex0);
                            cV+=J010;
                            if (J010<0.0) LOOP=0;
                        }
                        if (j<c-1)
                        {
                            NB++;
                            J011=fx1*(gy1*ez0-gz0*ey1)-fy1*(gx1*ez0-gz0*ex1)+fz0*(gx1*ey1-gy1*ex1);
                            cV+=J011;
                            if (J011<0.0) LOOP=0;
                        }
                    }
                }
                if (k<s-1)
                {
                    if (i>0)
                    {
                        if (j>0)
                        {
                            NB++;
                            J100=fx0*(gy0*ez1-gz1*ey0)-fy0*(gx0*ez1-gz1*ex0)+fz1*(gx0*ey0-gy0*ex0);
                            cV+=J100;
                            if (J100<0.0) LOOP=0;
                        }
                        if (j<c-1)
                        {
                            NB++;
                            J101=fx1*(gy0*ez1-gz1*ey0)-fy0*(gx1*ez1-gz1*ex1)+fz1*(gx1*ey0-gy0*ex1);
                            cV+=J101;
                            if (J101<0.0) LOOP=0;
                        }
                    }
                    if (i<r-1)
                    {
                        if (j>0)
                        {
                            NB++;
                            J110=fx0*(gy1*ez1-gz1*ey1)-fy1*(gx0*ez1-gz1*ex0)+fz1*(gx0*ey1-gy1*ex0);
                            cV+=J110;
                            if (J110<0.0) LOOP=0;
                        }
                        if (j<c-1)
                        {
                            NB++;
                            J111=fx1*(gy1*ez1-gz1*ey1)-fy1*(gx1*ez1-gz1*ex1)+fz1*(gx1*ey1-gy1*ex1);
                            cV+=J111;
                            if (J111<0.0) LOOP=0;
                        }
                    }
                }
                if (cTV>0.0)
                {
                    cV/=(float)NB;
                    
                    cres+=(cV-cTV)*(cV-cTV);
                }
                
                
                res+=0.5*cres;
            }
    if (LOOP>0)
        return res;
    else
        return -1.0;
}

/********************************************************************************************************************************/	

int VoxelVolumeMatchingGrads3D(float *f, float *g, float *e, float *TV, int r, int c, int s, float **df, float **dg, float **de)
{
    int i, j, k, rc=r*c, NB;
    int i0, i1, i2, i3, i4, i5, i6;
    float f0, f1, f2, f3, f4, f5, f6, cdf0, cdf1, cdf2, cdf3, cdf4, cdf5, cdf6;
    float g0, g1, g2, g3, g4, g5, g6, cdg0, cdg1, cdg2, cdg3, cdg4, cdg5, cdg6;
    float e0, e1, e2, e3, e4, e5, e6, cde0, cde1, cde2, cde3, cde4, cde5, cde6;
    float dJ000df0=0.0, dJ000df1=0.0, dJ000df2=0.0, dJ000df3=0.0, dJ000df4=0.0, dJ000df5=0.0, dJ000df6=0.0, dJ000dg0=0.0, dJ000dg1=0.0, dJ000dg2=0.0, dJ000dg3=0.0, dJ000dg4=0.0, dJ000dg5=0.0, dJ000dg6=0.0, dJ000de0=0.0, dJ000de1=0.0, dJ000de2=0.0, dJ000de3=0.0, dJ000de4=0.0, dJ000de5=0.0, dJ000de6=0.0;
    float dJ001df0=0.0, dJ001df1=0.0, dJ001df2=0.0, dJ001df3=0.0, dJ001df4=0.0, dJ001df5=0.0, dJ001df6=0.0, dJ001dg0=0.0, dJ001dg1=0.0, dJ001dg2=0.0, dJ001dg3=0.0, dJ001dg4=0.0, dJ001dg5=0.0, dJ001dg6=0.0, dJ001de0=0.0, dJ001de1=0.0, dJ001de2=0.0, dJ001de3=0.0, dJ001de4=0.0, dJ001de5=0.0, dJ001de6=0.0;
    float dJ010df0=0.0, dJ010df1=0.0, dJ010df2=0.0, dJ010df3=0.0, dJ010df4=0.0, dJ010df5=0.0, dJ010df6=0.0, dJ010dg0=0.0, dJ010dg1=0.0, dJ010dg2=0.0, dJ010dg3=0.0, dJ010dg4=0.0, dJ010dg5=0.0, dJ010dg6=0.0, dJ010de0=0.0, dJ010de1=0.0, dJ010de2=0.0, dJ010de3=0.0, dJ010de4=0.0, dJ010de5=0.0, dJ010de6=0.0;
    float dJ011df0=0.0, dJ011df1=0.0, dJ011df2=0.0, dJ011df3=0.0, dJ011df4=0.0, dJ011df5=0.0, dJ011df6=0.0, dJ011dg0=0.0, dJ011dg1=0.0, dJ011dg2=0.0, dJ011dg3=0.0, dJ011dg4=0.0, dJ011dg5=0.0, dJ011dg6=0.0, dJ011de0=0.0, dJ011de1=0.0, dJ011de2=0.0, dJ011de3=0.0, dJ011de4=0.0, dJ011de5=0.0, dJ011de6=0.0;
    float dJ100df0=0.0, dJ100df1=0.0, dJ100df2=0.0, dJ100df3=0.0, dJ100df4=0.0, dJ100df5=0.0, dJ100df6=0.0, dJ100dg0=0.0, dJ100dg1=0.0, dJ100dg2=0.0, dJ100dg3=0.0, dJ100dg4=0.0, dJ100dg5=0.0, dJ100dg6=0.0, dJ100de0=0.0, dJ100de1=0.0, dJ100de2=0.0, dJ100de3=0.0, dJ100de4=0.0, dJ100de5=0.0, dJ100de6=0.0;
    float dJ101df0=0.0, dJ101df1=0.0, dJ101df2=0.0, dJ101df3=0.0, dJ101df4=0.0, dJ101df5=0.0, dJ101df6=0.0, dJ101dg0=0.0, dJ101dg1=0.0, dJ101dg2=0.0, dJ101dg3=0.0, dJ101dg4=0.0, dJ101dg5=0.0, dJ101dg6=0.0, dJ101de0=0.0, dJ101de1=0.0, dJ101de2=0.0, dJ101de3=0.0, dJ101de4=0.0, dJ101de5=0.0, dJ101de6=0.0;
    float dJ110df0=0.0, dJ110df1=0.0, dJ110df2=0.0, dJ110df3=0.0, dJ110df4=0.0, dJ110df5=0.0, dJ110df6=0.0, dJ110dg0=0.0, dJ110dg1=0.0, dJ110dg2=0.0, dJ110dg3=0.0, dJ110dg4=0.0, dJ110dg5=0.0, dJ110dg6=0.0, dJ110de0=0.0, dJ110de1=0.0, dJ110de2=0.0, dJ110de3=0.0, dJ110de4=0.0, dJ110de5=0.0, dJ110de6=0.0;
    float dJ111df0=0.0, dJ111df1=0.0, dJ111df2=0.0, dJ111df3=0.0, dJ111df4=0.0, dJ111df5=0.0, dJ111df6=0.0, dJ111dg0=0.0, dJ111dg1=0.0, dJ111dg2=0.0, dJ111dg3=0.0, dJ111dg4=0.0, dJ111dg5=0.0, dJ111dg6=0.0, dJ111de0=0.0, dJ111de1=0.0, dJ111de2=0.0, dJ111de3=0.0, dJ111de4=0.0, dJ111de5=0.0, dJ111de6=0.0;
    float fx0, fx1, gx0, gx1, ex0, ex1;
    float fy0, fy1, gy0, gy1, ey0, ey1;
    float fz0, fz1, gz0, gz1, ez0, ez1;
    float J000, J001, J010, J011, J100, J101, J110, J111;
    int J000IS, J001IS, J010IS, J011IS, J100IS, J101IS, J110IS, J111IS;
    float cV, cTV, ErrComp, COND;
    
    for (k=0;k<s;k++)
        for (j=0;j<c;j++)
            for (i=0;i<r;i++)
            {
                i0=k*rc+j*r+i;
                cTV=TV[i0];
                f0=f[i0]; g0=g[i0]; e0=e[i0]; 
                if (j>0)
                {
                    i1=i0-r;
                    f1=f[i1]; g1=g[i1]; e1=e[i1]; 
                    fx0=f0-f1;
                    gx0=g0-g1;
                    ex0=e0-e1;
                }
                if (j<c-1)
                {
                    i2=i0+r;
                    f2=f[i2]; g2=g[i2]; e2=e[i2]; 
                    fx1=f2-f0;
                    gx1=g2-g0;
                    ex1=e2-e0;
                }
                if (i>0)
                {
                    i3=i0-1;
                    f3=f[i3]; g3=g[i3]; e3=e[i3]; 
                    fy0=f0-f3;
                    gy0=g0-g3;
                    ey0=e0-e3;
                }
                if (i<r-1)
                {
                    i4=i0+1;
                    f4=f[i4]; g4=g[i4]; e4=e[i4]; 
                    fy1=f4-f0;
                    gy1=g4-g0;
                    ey1=e4-e0;
                }
                if (k>0)
                {
                    i5=i0-rc;
                    f5=f[i5]; g5=g[i5]; e5=e[i5]; 
                    fz0=f0-f5;
                    gz0=g0-g5;
                    ez0=e0-e5;
                }
                if (k<s-1)
                {
                    i6=i0+rc;
                    f6=f[i6]; g6=g[i6]; e6=e[i6]; 
                    fz1=f6-f0;
                    gz1=g6-g0;
                    ez1=e6-e0;
                }
                
                cdf0=0.0; cdf1=0.0; cdf2=0.0; cdf3=0.0; cdf4=0.0; cdf5=0.0; cdf6=0.0; 
                cdg0=0.0; cdg1=0.0; cdg2=0.0; cdg3=0.0; cdg4=0.0; cdg5=0.0; cdg6=0.0; 
                cde0=0.0; cde1=0.0; cde2=0.0; cde3=0.0; cde4=0.0; cde5=0.0; cde6=0.0; 
                NB=0;
                cV=0.0;
                if (k>0)
                {
                    if (i>0)
                    {
                        if (j>0)
                        {
                            J000IS=1;
                            dJ000df0=((g0-g3)*(e0-e5)-(g0-g5)*(e0-e3)-(g0-g1)*(e0-e5)+(g0-g5)*(e0-e1)+(g0-g1)*(e0-e3)-(g0-g3)*(e0-e1));
                            dJ000df1=(-(g0-g3)*(e0-e5)+(g0-g5)*(e0-e3));
                            dJ000df3=((g0-g1)*(e0-e5)-(g0-g5)*(e0-e1));
                            dJ000df5=(-(g0-g1)*(e0-e3)+(g0-g3)*(e0-e1));
                            dJ000dg0=((f0-f1)*(-e5+e3)-(f0-f3)*(-e5+e1)+(f0-f5)*(-e3+e1));
                            dJ000dg1=(-(f0-f3)*(-e0+e5)+(f0-f5)*(-e0+e3));
                            dJ000dg3=((f0-f1)*(-e0+e5)+(f0-f5)*(e0-e1));
                            dJ000dg5=((f0-f1)*(e0-e3)-(f0-f3)*(e0-e1));
                            dJ000de0=((f0-f1)*(-g3+g5)-(f0-f3)*(-g1+g5)+(f0-f5)*(-g1+g3));
                            dJ000de1=(-(f0-f3)*(g0-g5)+(f0-f5)*(g0-g3));
                            dJ000de3=((f0-f1)*(g0-g5)+(f0-f5)*(-g0+g1));
                            dJ000de5=((f0-f1)*(-g0+g3)-(f0-f3)*(-g0+g1));
                            cdf0+=dJ000df0;
                            cdf1+=dJ000df1;
                            cdf3+=dJ000df3;
                            cdf5+=dJ000df5;
                            cdg0+=dJ000dg0;
                            cdg1+=dJ000dg1;
                            cdg3+=dJ000dg3;
                            cdg5+=dJ000dg5;
                            cde0+=dJ000de0;
                            cde1+=dJ000de1;
                            cde3+=dJ000de3;
                            cde5+=dJ000de5;                            
                            J000=fx0*(gy0*ez0-gz0*ey0)-fy0*(gx0*ez0-gz0*ex0)+fz0*(gx0*ey0-gy0*ex0);
                            cV+=J000;
                            NB++;
                        } else
                            J000IS=0;
                        if (j<c-1)
                        {
                            J001IS=1;
                            dJ001df0=(-(g0-g3)*(e0-e5)+(g0-g5)*(e0-e3)-(g2-g0)*(e0-e5)+(g0-g5)*(e2-e0)+(g2-g0)*(e0-e3)-(g0-g3)*(e2-e0));
                            dJ001df2=((g0-g3)*(e0-e5)-(g0-g5)*(e0-e3));
                            dJ001df3=((g2-g0)*(e0-e5)-(g0-g5)*(e2-e0));
                            dJ001df5=(-(g2-g0)*(e0-e3)+(g0-g3)*(e2-e0));
                            dJ001dg0=((f2-f0)*(-e5+e3)-(f0-f3)*(e5-e2)+(f0-f5)*(e3-e2));
                            dJ001dg2=(-(f0-f3)*(e0-e5)+(f0-f5)*(e0-e3));
                            dJ001dg3=((f2-f0)*(-e0+e5)+(f0-f5)*(e2-e0));
                            dJ001dg5=((f2-f0)*(e0-e3)-(f0-f3)*(e2-e0));
                            dJ001de0=((f2-f0)*(-g3+g5)-(f0-f3)*(g2-g5)+(f0-f5)*(g2-g3));
                            dJ001de2=(-(f0-f3)*(-g0+g5)+(f0-f5)*(-g0+g3));
                            dJ001de3=((f2-f0)*(g0-g5)+(f0-f5)*(-g2+g0));
                            dJ001de5=((f2-f0)*(-g0+g3)-(f0-f3)*(-g2+g0));
                            cdf0+=dJ001df0;
                            cdf2+=dJ001df2;
                            cdf3+=dJ001df3;
                            cdf5+=dJ001df5;
                            cdg0+=dJ001dg0;
                            cdg2+=dJ001dg2;
                            cdg3+=dJ001dg3;
                            cdg5+=dJ001dg5;
                            cde0+=dJ001de0;
                            cde2+=dJ001de2;
                            cde3+=dJ001de3;
                            cde5+=dJ001de5;
                            J001=fx1*(gy0*ez0-gz0*ey0)-fy0*(gx1*ez0-gz0*ex1)+fz0*(gx1*ey0-gy0*ex1);
                            cV+=J001;
                            NB++;
                        } else
                            J001IS=0;
                    }
                    if (i<r-1)
                    {
                        if (j>0)
                        {
                            J010IS=1;
                            dJ010df0=((g4-g0)*(e0-e5)-(g0-g5)*(e4-e0)+(g0-g1)*(e0-e5)-(g0-g5)*(e0-e1)+(g0-g1)*(e4-e0)-(g4-g0)*(e0-e1));
                            dJ010df1=(-(g4-g0)*(e0-e5)+(g0-g5)*(e4-e0));
                            dJ010df4=(-(g0-g1)*(e0-e5)+(g0-g5)*(e0-e1));
                            dJ010df5=(-(g0-g1)*(e4-e0)+(g4-g0)*(e0-e1));
                            dJ010dg0=((f0-f1)*(e5-e4)-(f4-f0)*(-e5+e1)+(f0-f5)*(e4-e1));
                            dJ010dg1=(-(f4-f0)*(-e0+e5)+(f0-f5)*(-e4+e0));
                            dJ010dg4=((f0-f1)*(e0-e5)+(f0-f5)*(-e0+e1));
                            dJ010dg5=((f0-f1)*(e4-e0)-(f4-f0)*(e0-e1));
                            dJ010de0=((f0-f1)*(g4-g5)-(f4-f0)*(-g1+g5)+(f0-f5)*(g1-g4));
                            dJ010de1=(-(f4-f0)*(g0-g5)+(f0-f5)*(g4-g0));
                            dJ010de4=((f0-f1)*(-g0+g5)+(f0-f5)*(g0-g1));
                            dJ010de5=((f0-f1)*(-g4+g0)-(f4-f0)*(-g0+g1));
                            cdf0+=dJ010df0;
                            cdf1+=dJ010df1;
                            cdf4+=dJ010df4;
                            cdf5+=dJ010df5;
                            cdg0+=dJ010dg0;
                            cdg1+=dJ010dg1;
                            cdg4+=dJ010dg4;
                            cdg5+=dJ010dg5;
                            cde0+=dJ010de0;
                            cde1+=dJ010de1;
                            cde4+=dJ010de4;
                            cde5+=dJ010de5;
                            J010=fx0*(gy1*ez0-gz0*ey1)-fy1*(gx0*ez0-gz0*ex0)+fz0*(gx0*ey1-gy1*ex0);
                            cV+=J010;
                            NB++;
                        } else
                            J010IS=0;
                        if (j<c-1)
                        {
                            J011IS=1;
                            dJ011df0=(-(g4-g0)*(e0-e5)+(g0-g5)*(e4-e0)+(g2-g0)*(e0-e5)-(g0-g5)*(e2-e0)+(g2-g0)*(e4-e0)-(g4-g0)*(e2-e0));
                            dJ011df2=((g4-g0)*(e0-e5)-(g0-g5)*(e4-e0));
                            dJ011df4=(-(g2-g0)*(e0-e5)+(g0-g5)*(e2-e0));
                            dJ011df5=(-(g2-g0)*(e4-e0)+(g4-g0)*(e2-e0));
                            dJ011dg0=((f2-f0)*(e5-e4)-(f4-f0)*(e5-e2)+(f0-f5)*(-e4+e2));
                            dJ011dg2=(-(f4-f0)*(e0-e5)+(f0-f5)*(e4-e0));
                            dJ011dg4=((f2-f0)*(e0-e5)+(f0-f5)*(-e2+e0));
                            dJ011dg5=((f2-f0)*(e4-e0)-(f4-f0)*(e2-e0));
                            dJ011de0=((f2-f0)*(g4-g5)-(f4-f0)*(g2-g5)+(f0-f5)*(-g2+g4));
                            dJ011de2=(-(f4-f0)*(-g0+g5)+(f0-f5)*(-g4+g0));
                            dJ011de4=((f2-f0)*(-g0+g5)+(f0-f5)*(g2-g0));
                            dJ011de5=((f2-f0)*(-g4+g0)-(f4-f0)*(-g2+g0));
                            cdf0+=dJ011df0;
                            cdf2+=dJ011df2;
                            cdf4+=dJ011df4;
                            cdf5+=dJ011df5;
                            cdg0+=dJ011dg0;
                            cdg2+=dJ011dg2;
                            cdg4+=dJ011dg4;
                            cdg5+=dJ011dg5;
                            cde0+=dJ011de0;
                            cde2+=dJ011de2;
                            cde4+=dJ011de4;
                            cde5+=dJ011de5;
                            J011=fx1*(gy1*ez0-gz0*ey1)-fy1*(gx1*ez0-gz0*ex1)+fz0*(gx1*ey1-gy1*ex1);
                            cV+=J011;
                            NB++;
                        } else
                            J011IS=0;
                    }
                }
                if (k<s-1)
                {
                    if (i>0)
                    {
                        if (j>0)
                        {
                            J100IS=1;
                            dJ100df0=((g0-g3)*(e6-e0)-(g6-g0)*(e0-e3)-(g0-g1)*(e6-e0)+(g6-g0)*(e0-e1)-(g0-g1)*(e0-e3)+(g0-g3)*(e0-e1));
                            dJ100df1=(-(g0-g3)*(e6-e0)+(g6-g0)*(e0-e3));
                            dJ100df3=((g0-g1)*(e6-e0)-(g6-g0)*(e0-e1));
                            dJ100df6=((g0-g1)*(e0-e3)-(g0-g3)*(e0-e1));
                            dJ100dg0=((f0-f1)*(e6-e3)-(f0-f3)*(e6-e1)+(f6-f0)*(-e3+e1));
                            dJ100dg1=(-(f0-f3)*(-e6+e0)+(f6-f0)*(-e0+e3));
                            dJ100dg3=((f0-f1)*(-e6+e0)+(f6-f0)*(e0-e1));
                            dJ100dg6=((f0-f1)*(-e0+e3)-(f0-f3)*(-e0+e1));
                            dJ100de0=((f0-f1)*(g3-g6)-(f0-f3)*(g1-g6)+(f6-f0)*(-g1+g3));
                            dJ100de1=(-(f0-f3)*(g6-g0)+(f6-f0)*(g0-g3));
                            dJ100de3=((f0-f1)*(g6-g0)+(f6-f0)*(-g0+g1));
                            dJ100de6=((f0-f1)*(g0-g3)-(f0-f3)*(g0-g1));
                            cdf0+=dJ100df0;
                            cdf1+=dJ100df1;
                            cdf3+=dJ100df3;
                            cdf6+=dJ100df6;
                            cdg0+=dJ100dg0;
                            cdg1+=dJ100dg1;
                            cdg3+=dJ100dg3;
                            cdg6+=dJ100dg6;
                            cde0+=dJ100de0;
                            cde1+=dJ100de1;
                            cde3+=dJ100de3;
                            cde6+=dJ100de6;
                            J100=fx0*(gy0*ez1-gz1*ey0)-fy0*(gx0*ez1-gz1*ex0)+fz1*(gx0*ey0-gy0*ex0);
                            cV+=J100;
                            NB++;
                        } else
                            J100IS=0;
                        if (j<c-1)
                        {
                            J101IS=1;
                            dJ101df0=(-(g0-g3)*(e6-e0)+(g6-g0)*(e0-e3)-(g2-g0)*(e6-e0)+(g6-g0)*(e2-e0)-(g2-g0)*(e0-e3)+(g0-g3)*(e2-e0));
                            dJ101df2=((g0-g3)*(e6-e0)-(g6-g0)*(e0-e3));
                            dJ101df3=((g2-g0)*(e6-e0)-(g6-g0)*(e2-e0));
                            dJ101df6=((g2-g0)*(e0-e3)-(g0-g3)*(e2-e0));
                            dJ101dg0=((f2-f0)*(e6-e3)-(f0-f3)*(-e6+e2)+(f6-f0)*(e3-e2));
                            dJ101dg2=(-(f0-f3)*(e6-e0)+(f6-f0)*(e0-e3));
                            dJ101dg3=((f2-f0)*(-e6+e0)+(f6-f0)*(e2-e0));
                            dJ101dg6=((f2-f0)*(-e0+e3)-(f0-f3)*(-e2+e0));
                            dJ101de0=((f2-f0)*(g3-g6)-(f0-f3)*(-g2+g6)+(f6-f0)*(g2-g3));
                            dJ101de2=(-(f0-f3)*(-g6+g0)+(f6-f0)*(-g0+g3));
                            dJ101de3=((f2-f0)*(g6-g0)+(f6-f0)*(-g2+g0));
                            dJ101de6=((f2-f0)*(g0-g3)-(f0-f3)*(g2-g0));
                            cdf0+=dJ101df0;
                            cdf2+=dJ101df2;
                            cdf3+=dJ101df3;
                            cdf6+=dJ101df6;
                            cdg0+=dJ101dg0;
                            cdg2+=dJ101dg2;
                            cdg3+=dJ101dg3;
                            cdg6+=dJ101dg6;
                            cde0+=dJ101de0;
                            cde2+=dJ101de2;
                            cde3+=dJ101de3;
                            cde6+=dJ101de6;
                            J101=fx1*(gy0*ez1-gz1*ey0)-fy0*(gx1*ez1-gz1*ex1)+fz1*(gx1*ey0-gy0*ex1);
                            cV+=J101;
                            NB++;
                        } else
                            J101IS=0;
                    }
                    if (i<r-1)
                    {
                        if (j>0)
                        {
                            J110IS=1;
                            dJ110df0=((g4-g0)*(e6-e0)-(g6-g0)*(e4-e0)+(g0-g1)*(e6-e0)-(g6-g0)*(e0-e1)-(g0-g1)*(e4-e0)+(g4-g0)*(e0-e1));
                            dJ110df1=(-(g4-g0)*(e6-e0)+(g6-g0)*(e4-e0));
                            dJ110df4=(-(g0-g1)*(e6-e0)+(g6-g0)*(e0-e1));
                            dJ110df6=((g0-g1)*(e4-e0)-(g4-g0)*(e0-e1));
                            dJ110dg0=((f0-f1)*(-e6+e4)-(f4-f0)*(e6-e1)+(f6-f0)*(e4-e1));
                            dJ110dg1=(-(f4-f0)*(-e6+e0)+(f6-f0)*(-e4+e0));
                            dJ110dg4=((f0-f1)*(e6-e0)+(f6-f0)*(-e0+e1));
                            dJ110dg6=((f0-f1)*(-e4+e0)-(f4-f0)*(-e0+e1));
                            dJ110de0=((f0-f1)*(-g4+g6)-(f4-f0)*(g1-g6)+(f6-f0)*(g1-g4));
                            dJ110de1=(-(f4-f0)*(g6-g0)+(f6-f0)*(g4-g0));
                            dJ110de4=((f0-f1)*(-g6+g0)+(f6-f0)*(g0-g1));
                            dJ110de6=((f0-f1)*(g4-g0)-(f4-f0)*(g0-g1));
                            cdf0+=dJ110df0;
                            cdf1+=dJ110df1;
                            cdf4+=dJ110df4;
                            cdf6+=dJ110df6;
                            cdg0+=dJ110dg0;
                            cdg1+=dJ110dg1;
                            cdg4+=dJ110dg4;
                            cdg6+=dJ110dg6;
                            cde0+=dJ110de0;
                            cde1+=dJ110de1;
                            cde4+=dJ110de4;
                            cde6+=dJ110de6;
                            J110=fx0*(gy1*ez1-gz1*ey1)-fy1*(gx0*ez1-gz1*ex0)+fz1*(gx0*ey1-gy1*ex0);
                            cV+=J110;
                            NB++;
                        } else
                            J110IS=0;
                        if (j<c-1)
                        {
                            J111IS=1;
                            dJ111df0=(-(g4-g0)*(e6-e0)+(g6-g0)*(e4-e0)+(g2-g0)*(e6-e0)-(g6-g0)*(e2-e0)-(g2-g0)*(e4-e0)+(g4-g0)*(e2-e0));
                            dJ111df2=((g4-g0)*(e6-e0)-(g6-g0)*(e4-e0));
                            dJ111df4=(-(g2-g0)*(e6-e0)+(g6-g0)*(e2-e0));
                            dJ111df6=((g2-g0)*(e4-e0)-(g4-g0)*(e2-e0));
                            dJ111dg0=((f2-f0)*(-e6+e4)-(f4-f0)*(-e6+e2)+(f6-f0)*(-e4+e2));
                            dJ111dg2=(-(f4-f0)*(e6-e0)+(f6-f0)*(e4-e0));
                            dJ111dg4=((f2-f0)*(e6-e0)+(f6-f0)*(-e2+e0));
                            dJ111dg6=((f2-f0)*(-e4+e0)-(f4-f0)*(-e2+e0));
                            dJ111de0=((f2-f0)*(-g4+g6)-(f4-f0)*(-g2+g6)+(f6-f0)*(-g2+g4));
                            dJ111de2=(-(f4-f0)*(-g6+g0)+(f6-f0)*(-g4+g0));
                            dJ111de4=((f2-f0)*(-g6+g0)+(f6-f0)*(g2-g0));
                            dJ111de6=((f2-f0)*(g4-g0)-(f4-f0)*(g2-g0));
                            cdf0+=dJ111df0;
                            cdf2+=dJ111df2;
                            cdf4+=dJ111df4;
                            cdf6+=dJ111df6;
                            cdg0+=dJ111dg0;
                            cdg2+=dJ111dg2;
                            cdg4+=dJ111dg4;
                            cdg6+=dJ111dg6;
                            cde0+=dJ111de0;
                            cde2+=dJ111de2;
                            cde4+=dJ111de4;
                            cde6+=dJ111de6;
                            J111=fx1*(gy1*ez1-gz1*ey1)-fy1*(gx1*ez1-gz1*ex1)+fz1*(gx1*ey1-gy1*ex1);
                            cV+=J111;
                            NB++;
                        } else
                            J111IS=0;
                    }
                }
                cV/=(float)NB;
                
                if (cTV>0.0)
                {
                    ErrComp=(cV-cTV)/((float)NB);
                    
                    (*df)[i0]+=ErrComp*cdf0;
                    (*dg)[i0]+=ErrComp*cdg0;
                    (*de)[i0]+=ErrComp*cde0;
                    if (j>0)
                    {
                        (*df)[i1]+=ErrComp*cdf1;
                        (*dg)[i1]+=ErrComp*cdg1;
                        (*de)[i1]+=ErrComp*cde1;
                    }
                    if (j<c-1)
                    {
                        (*df)[i2]+=ErrComp*cdf2;
                        (*dg)[i2]+=ErrComp*cdg2;
                        (*de)[i2]+=ErrComp*cde2;
                    }
                    if (i>0)
                    {
                        (*df)[i3]+=ErrComp*cdf3;
                        (*dg)[i3]+=ErrComp*cdg3;
                        (*de)[i3]+=ErrComp*cde3;
                    }
                    if (i<r-1)
                    {
                        (*df)[i4]+=ErrComp*cdf4;
                        (*dg)[i4]+=ErrComp*cdg4;
                        (*de)[i4]+=ErrComp*cde4;
                    }
                    if (k>0)
                    {
                        (*df)[i5]+=ErrComp*cdf5;
                        (*dg)[i5]+=ErrComp*cdg5;
                        (*de)[i5]+=ErrComp*cde5;
                    }
                    if (k<s-1)
                    {
                        (*df)[i6]+=ErrComp*cdf6;
                        (*dg)[i6]+=ErrComp*cdg6;
                        (*de)[i6]+=ErrComp*cde6;
                    }
                }
            }
    return 1;
}

/********************************************************************************************************************************/	

int VoxelVolumeMatching3D(float **f, float **g, float **e, float *TV, int r, int c, int s, int maxitnum, int norestriction, unsigned char *RestrictionMap, float Jmin, float lambda)
{
    int i, j, k, ind, rc=r*c, rcs=r*c*s, itnum;
    float *nf, *ng, *ne;
    float *df, *dg, *de;
    float *H, Hmark, nH, alpha=1.0, VMC, TC;
    
    H=(float *)malloc(sizeof(float)*(maxitnum+1));
    nf=(float *)malloc(sizeof(float)*rcs);
    ng=(float *)malloc(sizeof(float)*rcs);
    ne=(float *)malloc(sizeof(float)*rcs);
    df=(float *)malloc(sizeof(float)*rcs);
    dg=(float *)malloc(sizeof(float)*rcs);
    de=(float *)malloc(sizeof(float)*rcs);
    
    for (i=0;i<rcs;i++)
    {
        df[i]=0.0;
        dg[i]=0.0;
        de[i]=0.0;
    }
    
    VMC=VoxelVolumeMatchingCost3D(*f,*g,*e,TV,r,c,s);
    TC=TopologyCost3D(*f,*g,*e,r,c,s,Jmin,lambda);
    H[0]=VMC+TC;
    
    itnum=0;
    while ((itnum<maxitnum)&&(alpha>0.000000001))
    {
        Hmark=H[itnum];
        VoxelVolumeMatchingGrads3D((*f),(*g),(*e),TV,r,c,s,&df,&dg,&de);
        TopologyGrads3D((*f),(*g),(*e),r,c,s,Jmin,lambda,&df,&dg,&de);
        for (i=0;i<rcs;i++)
        {
            nf[i]=(*f)[i]-alpha*df[i];
            ng[i]=(*g)[i]-alpha*dg[i];
            ne[i]=(*e)[i]-alpha*de[i];
        }
        VMC=VoxelVolumeMatchingCost3D(nf,ng,ne,TV,r,c,s);
        if (VMC>-.5) TC=TopologyCost3D(nf,ng,ne,r,c,s,Jmin,lambda); else TC=1e300;
        nH=VMC+TC;
        if (nH<Hmark)
        {
            if (itnum==0)
            {
                while (nH<Hmark)
                {
                    if (VERBOSITY) {printf("/"); fflush(stdout);}
                    Hmark=nH;
                    alpha*=2.0;
                    for (i=0;i<rcs;i++)
                    {
                        nf[i]=(*f)[i]-alpha*df[i];
                        ng[i]=(*g)[i]-alpha*dg[i];
                        ne[i]=(*e)[i]-alpha*de[i];
                    }
                    VMC=VoxelVolumeMatchingCost3D(nf,ng,ne,TV,r,c,s);
                    if (VMC>-.5) TC=TopologyCost3D(nf,ng,ne,r,c,s,Jmin,lambda); else TC=1e300;
                    nH=VMC+TC;
                }
                if (VERBOSITY) {printf("\\"); fflush(stdout);}
                alpha*=0.5;
                for (i=0;i<rcs;i++)
                {
                    nf[i]=(*f)[i]-alpha*df[i];
                    ng[i]=(*g)[i]-alpha*dg[i];
                    ne[i]=(*e)[i]-alpha*de[i];
                }
                H[itnum+1]=Hmark;
            } else
            {
                H[itnum+1]=nH;
                alpha*=1.0905;
            }
        } else
        {
            while ((nH>Hmark)&&(alpha>.000000001))
            {
                if (VERBOSITY) {printf("\\"); fflush(stdout);}
                alpha*=0.5;
                for (i=0;i<rcs;i++)
                {
                    nf[i]=(*f)[i]-alpha*df[i];
                    ng[i]=(*g)[i]-alpha*dg[i];
                    ne[i]=(*e)[i]-alpha*de[i];
                }
                VMC=VoxelVolumeMatchingCost3D(nf,ng,ne,TV,r,c,s);
                if (VMC>-.5) TC=TopologyCost3D(nf,ng,ne,r,c,s,Jmin,lambda); else TC=1e300;
                nH=VMC+TC;
            }
            H[itnum+1]=nH;
        }
        if (alpha>.000000001)
        {
            if (norestriction)
                for (i=0;i<rcs;i++)
                {
                    (*f)[i]=nf[i];
                    (*g)[i]=ng[i];
                    (*e)[i]=ne[i];
                    df[i]=0.0;
                    dg[i]=0.0;
                    de[i]=0.0;
                }
            else
                for (i=0;i<rcs;i++)
                {
                    if (RestrictionMap[i])
                    {
                        (*f)[i]=nf[i];
                        (*g)[i]=ng[i];
                        (*e)[i]=ne[i];
                    }
                    df[i]=0.0;
                    dg[i]=0.0;
                    de[i]=0.0;
                }
            itnum++;
        }
        if (VERBOSITY) {printf("."); fflush(stdout);}
    }
    if (itnum<maxitnum) fprintf(stderr,"(convergence at iteration %d.)",itnum);
    
    if (VERBOSITY)
    {
        printf("\n");
        for (i=0;i<maxitnum+1;i++)
        {
            printf("%12.4f\n",H[i]); 
            fflush(stdout);
        }
    }
    /* iterations have been exhaused, freeing memory and returning*/
    free(nf);
    free(ng);
    free(ne);
    free(df);
    free(dg);
    free(de);
    free(H);
}

/********************************************************************************************************************************/	

void Usage(char *programName)
{
    int i,n=0;
    char *emptyspace;
    
    while (programName[n]!='\0')
        n++;
    emptyspace=(char *)malloc(sizeof(char)*n);
    for (i=0;i<n;i++)
        emptyspace[i]=' ';
    emptyspace[n]='\0';
	fprintf(stderr,"%s usage:\n",programName);
	fprintf(stderr,"%s finds the deformation fields that agree with the prescribed values for cubic volumes to the extent possible. ",programName);
    fprintf(stderr,"Specifically, it reads an initial deformation field and the desired set of volumes, and employs a ");
    fprintf(stderr,"gradient descent algorithm to minimize the squared error between the prescribed set of volumes to the average of corner ");
    fprintf(stderr,"Jacobians of the deformation field at every voxel. Note that it reads one desired volume per voxel; ");
    fprintf(stderr,"thus, for a r x c x s deformation field, it expects to read r*c*s volume values. ");
    fprintf(stderr,"It assumes the deformation field is defined on an isotropic grid with equal spacing in x, y, and z directions. %s takes the following options:\n",programName);
    fprintf(stderr,"%s -d <displacement field> \n",emptyspace);
    fprintf(stderr,"%s -v <prescribed volumes> \n",emptyspace);
    fprintf(stderr,"%s -o <output file name [defaultoutput.dat]> \n",emptyspace);
    fprintf(stderr,"%s -r <number of rows (y dim) [256]> \n",emptyspace);
    fprintf(stderr,"%s -c <number of columns (x dim) [256]> \n",emptyspace);
    fprintf(stderr,"%s -s <number of slices (z dim) [198]> \n",emptyspace);
    fprintf(stderr,"%s -I <number of iterations [1000]> \n",emptyspace);
    fprintf(stderr,"%s -R <restrictions file (char)> \n",emptyspace);
    fprintf(stderr,"%s -L <weight of the topology preservation constraint (1.0)> \n",emptyspace);
    fprintf(stderr,"%s -J <minimum allowed Jacobian value (0.001)> \n",emptyspace);
    fprintf(stderr,"%s -V <verbosity [1]> \n",emptyspace);
    
    free(emptyspace);
}

/********************************************************************************************************************************/	

/* returns the index of the first argument that is not an option; i.e.
   does not start with a dash or a slash
*/
int HandleOptions(int argc,char *argv[])
{
	int i,firstnonoption=0;
	float tau,popt;

	for (i=1; i< argc;i++) {
/*		if (argv[i][0] == '/' || argv[i][0] == '-') {*/
		if (argv[i][0] == '-') {
			switch (argv[i][1]) {
				/* An argument -? means help is requested */
				case '?':
					Usage(argv[0]);
					break;
/*				case 'h':
				case 'H':
					if (!stricmp(argv[i]+1,"help")) {
						Usage(argv[0]);
						break;
					}
					/* If the option -h means anything else
					 * in your application add code here
					 * Note: this falls through to the default
					 * to print an "unknow option" message
					*/
				/* add your option switches here */
				case 'd':
					strcpy((char *)&Deformationfieldfilename,argv[i+1]);
                    NODEFORMATION=0;
					break;
				case 'v':
					strcpy((char *)&Volumefilename,argv[i+1]);
					break;
				case 'R':
					strcpy((char *)&Restrictionfilename,argv[i+1]);
                    NORESTRICTION=0;
					break;
				case 'o':
					strcpy((char *)&Outputfilename,argv[i+1]);
					NOOUTFILE=0;
					break;
				case 'r':
					r=atoi(argv[i+1]);
                    if (r<2)
                    {
                        fprintf(stderr,"number of rows too small (%d).\n",r);
                        fprintf(stderr,"bailing out ..\n");
                        return 0;
                    }
					break;
				case 'c':
					c=atoi(argv[i+1]);
                    if (c<2)
                    {
                        fprintf(stderr,"number of columns too small (%d).\n",c);
                        fprintf(stderr,"bailing out ..\n");
                        return 0;
                    }
					break;
				case 's':
					s=atoi(argv[i+1]);
                    if (s<2)
                    {
                        fprintf(stderr,"number of slices too small (%d).\n",s);
                        fprintf(stderr,"bailing out ..\n");
                        return 0;
                    }
					break;
				case 'I':
					MAXITNUM=atoi(argv[i+1]);
					break;
				case 'D':
					DF=0;
					break;
				case 'L':
					LAMBDA=atof(argv[i+1]);
					break;
				case 'J':
					JMIN=atof(argv[i+1]);
					break;
				case 'V':
					VERBOSITY=atoi(argv[i+1]);
					break;
				default:
					fprintf(stderr,"unknown option %s\n",argv[i]);
					break;
			}
		}
	}
	return 1;
}



/********************************************************************************************************************************/	

int main(int argc,char *argv[])
{
    float *f, *g, *e, *TV;
    unsigned char *RestrictionMap, ctemp;
	int i,j,k,ind,L;
	float temp;
    int correction;
    int mindim;
    FILE *Volumefile, *Restrictionfile;
	
	if (argc == 1) {
		/* If no arguments we call the Usage routine and exit */
		Usage(argv[0]);
		return 1;
	}
	/* handle the program options */
	if (HandleOptions(argc,argv))
    {
        L=r*c*s;
        if (NODEFORMATION)
        {
            f=(float *)calloc(L,sizeof(float));
            g=(float *)calloc(L,sizeof(float));
            e=(float *)calloc(L,sizeof(float));
            ind=0;
            for (k=1;k<s+1;k++)
                for (j=1;j<c+1;j++)
                    for (i=1;i<r+1;i++)
                    {
                        f[ind]=(float)j;
                        g[ind]=(float)i;
                        e[ind]=(float)k;
                        ind++;
                    }
        } else
            if (!LoadDeformationFieldCartesianComponents3D(Deformationfieldfilename,&f,&g,&e,r,c,s))
            {
                fprintf(stderr,"cannot open the file %s for read\n",Deformationfieldfilename);
                fprintf(stderr,"starting from a null deformation\n");
                f=(float *)calloc(r*c*s,sizeof(float));
                g=(float *)calloc(r*c*s,sizeof(float));
                e=(float *)calloc(r*c*s,sizeof(float));
                ind=0;
                for (k=1;k<s+1;k++)
                    for (j=1;j<c+1;j++)
                        for (i=1;i<r+1;i++)
                        {
                            f[ind]=(float)j;
                            g[ind]=(float)i;
                            e[ind]=(float)k;
                            ind++;
                        }
                }
        /* Load the Jacobians*/
        if ((Volumefile=fopen(Volumefilename,"rb"))==NULL)
        {
            fprintf(stderr,"Cannot open the volume file %s for read.\n",Volumefilename);
            fprintf(stderr,"no action performed.\n");
        } else
        {
            if (NORESTRICTION==0)
            {
                if ((Restrictionfile=fopen(Restrictionfilename,"rb"))==NULL)
                {
                    fprintf(stderr,"Cannot open the restriction map file %s for read.\n",Restrictionfilename);
                    fprintf(stderr,"will assume no restrictions.\n");
                    NORESTRICTION=1;
                } else
                {
                    RestrictionMap=(unsigned char *)malloc(sizeof(unsigned char)*r*c*s);
                    for (i=0;i<r*c*s;i++)
                    {
                        fread(&ctemp,sizeof(char),1,Restrictionfile);
                        RestrictionMap[i]=(ctemp>0);
                    }
                }
            }
            TV=(float *)malloc(sizeof(float)*L);
            for (i=0;i<L;i++)
            {
                fread(&temp,sizeof(float),1,Volumefile);
                TV[i]=temp;
            }
            fclose(Volumefile);
            VoxelVolumeMatching3D(&f,&g,&e,TV,r,c,s,MAXITNUM,NORESTRICTION,RestrictionMap,JMIN,LAMBDA);
       		
            if (DF)
            {
                if (NOOUTFILE)
                	SaveDeformationFieldCartesianComponents3D(Defaultoutputfilename,f,g,e,r,c,s);
   	            else
    		        SaveDeformationFieldCartesianComponents3D(Outputfilename,f,g,e,r,c,s);
            } else
            {
                ind=0;
                for (k=1;k<s+1;k++)
                    for (j=1;j<c+1;j++)
                        for (i=1;i<r+1;i++)
                        {
                            f[ind]-=(float)j;
                            g[ind]-=(float)i;
                            e[ind]-=(float)k;
                            ind++;
                        }
                
                if (NOOUTFILE)
                	SaveDeformationFieldCartesianComponents3D(Defaultoutputfilename,f,g,e,r,c,s);
   	            else
    		        SaveDeformationFieldCartesianComponents3D(Outputfilename,f,g,e,r,c,s);
            }
       		/* freeing the emmory allocated to f, g, and e */
            free(f);
            free(g);
            free(e);
            free(TV);
            free(RestrictionMap);
            return 1;
        }
    }
}


