/********************************************************************************************************************************/	
/* Disclaimer
/* This code is provided as is, without any guarantees. 
/* Comments and bug reports should be forwarded to bilge@rad.upenn.edu.
/* WarpImage3D -- Copyright Bilge Karacali 2004
/********************************************************************************************************************************/	

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <dfieldio.h>

/********************************************************************************************************************************/	

/* Predefined constants */
#define PI 3.14159265358979
#define rand_float() (((float)rand())/((float)RAND_MAX))

/********************************************************************************************************************************/	

/* Global Variables */
int R=256;
int C=256;
int S=198;
char DeformationFileName[300];
char ImageFileName[300];
char OutputFileName[300];
char IF='b';
char OF='b';
int NOOUTFILE=1;
float T=3;
int DF=0;
int VOLUMEDISTRIBUTION=0;
int INTERPOLATE=0;

/********************************************************************************************************************************/	

/* Function Declarations */
/* void SolveCubic(); */
void Usage();
int HandleOptions();
int WarpImage3D(float *, float *, float *, float *, int, int, int, float **, int, int);
int WarpGrayImage3D(float *, float *, float *, float *, int, int, int, float **, int, int);
int ReadImage3D(char *, float **, char);
int WriteImage3D(char *, float *, char, int);

/********************************************************************************************************************************/	
/* Function Definitions */

int WarpImage3D(float *f, float *g, float *e, float *I, int r, int c, int s, float **WI, int K, int interpolate)
{
    int i, j, k, rc=r*c, ind;
    int i000, i001, i010, i011, i100, i101, i110, i111;
    int xind, yind,zind, bi, bj, bk;
    float x, y, z, cf, cg, ce, cI, cw, NC, hx, hy, hz;
    
    for (k=0;k<s-1;k++)
        for (j=0;j<c-1;j++)
            for (i=0;i<r-1;i++)
            {
                i000=rc*k+r*j+i;
                i001=i000+r;
                i010=i000+1;
                i011=i001+1;
                i100=i000+rc;
                i101=i100+r;
                i110=i100+1;
                i111=i101+1;
                if ((I[i000]>1e-10)||(I[i001]>1e-10)||(I[i010]>1e-10)||(I[i011]>1e-10)||(I[i100]>1e-10)||(I[i101]>1e-10)||(I[i110]>1e-10)||(I[i111]>1e-10))
                    for (xind=1;xind<2*K;xind+=2)
                    {
                        x=((float)xind)/((float)2*K);
                        for (yind=1;yind<2*K;yind+=2)
                        {
                            y=((float)yind)/((float)2*K);
                            for (zind=1;zind<2*K;zind+=2)
                            {
                                z=((float)zind)/((float)2*K);
                                cf=x+(1.0-z)*((1.0-y)*((1.0-x)*f[i000]+x*f[i001])+y*((1.0-x)*f[i010]+x*f[i011]))+z*((1.0-y)*((1.0-x)*f[i100]+x*f[i101])+y*((1.0-x)*f[i110]+x*f[i111]));
                                cg=y+(1.0-z)*((1.0-y)*((1.0-x)*g[i000]+x*g[i001])+y*((1.0-x)*g[i010]+x*g[i011]))+z*((1.0-y)*((1.0-x)*g[i100]+x*g[i101])+y*((1.0-x)*g[i110]+x*g[i111]));
                                ce=z+(1.0-z)*((1.0-y)*((1.0-x)*e[i000]+x*e[i001])+y*((1.0-x)*e[i010]+x*e[i011]))+z*((1.0-y)*((1.0-x)*e[i100]+x*e[i101])+y*((1.0-x)*e[i110]+x*e[i111]));
                                /*cI= ( (z<.5) ? ( (y<.5) ? ( (x<.5) ? I[i000] : I[i001] ) : ( (x<.5) ? I[i010] : I[i011] ) ) : ( (y<.5) ? ( (x<.5) ? I[i100] : I[i101] ) : ( (x<.5) ? I[i110] : I[i111] ) ) );*/
                                hx=cf-floor(cf);
                                hy=cg-floor(cg);
                                hz=ce-floor(ce);
                                bj=((int)floor(cf))+j;
                                bi=((int)floor(cg))+i;
                                bk=((int)floor(ce))+k;
                                if (interpolate)
                                {
                                    cI=(1.0-z)*((1.0-y)*((1.0-x)*I[i000]+x*I[i001])+y*((1.0-x)*I[i010]+x*I[i011]))+z*((1.0-y)*((1.0-x)*I[i100]+x*I[i101])+y*((1.0-x)*I[i110]+x*I[i111]));
                                    if ((bk>=0)&&(bk<s))
                                    {
                                        if ((bj>=0)&&(bj<c))
                                        {
                                            if ((bi>=0)&&(bi<r))
                                            {
                                                cw=(1.0-hx)*(1.0-hy)*(1.0-hz);
                                                ind=rc*bk+r*bj+bi;
                                                (*WI)[ind]+=cw*cI;
                                            }
                                            if ((bi>=-1)&&(bi<r-1))
                                            {
                                                cw=(1.0-hx)*hy*(1.0-hz);
                                                ind=rc*bk+r*bj+bi+1;
                                                (*WI)[ind]+=cw*cI;
                                            }
                                        }
                                        if ((bj>=-1)&&(bj<c-1))
                                        {
                                            if ((bi>=0)&&(bi<r))
                                            {
                                                cw=hx*(1.0-hy)*(1.0-hz);
                                                ind=rc*bk+r*(bj+1)+bi;
                                                (*WI)[ind]+=cw*cI;
                                            }
                                            if ((bi>=-1)&&(bi<r-1))
                                            {
                                                cw=hx*hy*(1.0-hz);
                                                ind=rc*bk+r*(bj+1)+bi+1;
                                                (*WI)[ind]+=cw*cI;
                                            }
                                        }
                                    }
                                    if ((bk>=-1)&&(bk<s-1))
                                    {
                                        if ((bj>=0)&&(bj<c))
                                        {
                                            if ((bi>=0)&&(bi<r))
                                            {
                                                cw=(1.0-hx)*(1.0-hy)*hz;
                                                ind=rc*(bk+1)+r*bj+bi;
                                                (*WI)[ind]+=cw*cI;
                                            }
                                            if ((bi>=-1)&&(bi<r-1))
                                            {
                                                cw=(1.0-hx)*hy*hz;
                                                ind=rc*(bk+1)+r*bj+bi+1;
                                                (*WI)[ind]+=cw*cI;
                                            }
                                        }
                                        if ((bj>=-1)&&(bj<c-1))
                                        {
                                            if ((bi>=0)&&(bi<r))
                                            {
                                                cw=hx*(1.0-hy)*hz;
                                                ind=rc*(bk+1)+r*(bj+1)+bi;
                                                (*WI)[ind]+=cw*cI;
                                            }
                                            if ((bi>=-1)&&(bi<r-1))
                                            {
                                                cw=hx*hy*hz;
                                                ind=rc*(bk+1)+r*(bj+1)+bi+1;
                                                (*WI)[ind]+=cw*cI;
                                            }
                                        }
                                    }
                                }
                                else
                                {
                                cI= ( (z<.5) ? ( (y<.5) ? ( (x<.5) ? I[i000] : I[i001] ) : ( (x<.5) ? I[i010] : I[i011] ) ) : ( (y<.5) ? ( (x<.5) ? I[i100] : I[i101] ) : ( (x<.5) ? I[i110] : I[i111] ) ) );
                                    if (hz>=.5)
                                        bk++;
                                    if (hx>=.5)
                                        bj++;
                                    if (hy>=.5)
                                        bi++;
                                    if ((bi>=0)&(bi<r)&(bj>=0)&(bj<c)&(bk>=0)&(bk<s))
                                    {
                                        ind=rc*bk+r*bj+bi;
                                        (*WI)[ind]+=cI;
                                    }
                                }
                            }
                        }
                    }
            }
    NC=1.0/((float) (K*K*K));
    ind=0;
    for (k=0;k<s;k++)
        for (j=0;j<c;j++)
            for (i=0;i<r;i++)
            {
                (*WI)[ind]*=NC;
                ind++;
            }
    return 1;
}

/********************************************************************************************************************************/	

int WarpGrayImage3D(float *f, float *g, float *e, float *I, int r, int c, int s, float **WI, int K, int interpolate)
{
    int i, j, k, rc=r*c, ind;
    int i000, i001, i010, i011, i100, i101, i110, i111;
    int xind, yind,zind, bi, bj, bk;
    float x, y, z, cf, cg, ce, cI, hx, hy, hz;
    float *ws, cw;
    
    ws=(float *)calloc(r*c*s,sizeof(float));
    
    for (k=0;k<s-1;k++)
        for (j=0;j<c-1;j++)
            for (i=0;i<r-1;i++)
            {
                i000=rc*k+r*j+i;
                i001=i000+r;
                i010=i000+1;
                i011=i001+1;
                i100=i000+rc;
                i101=i100+r;
                i110=i100+1;
                i111=i101+1;
                if ((I[i000]>1e-10)||(I[i001]>1e-10)||(I[i010]>1e-10)||(I[i011]>1e-10)||(I[i100]>1e-10)||(I[i101]>1e-10)||(I[i110]>1e-10)||(I[i111]>1e-10))
                    for (xind=1;xind<2*K;xind+=2)
                    {
                        x=((float)xind)/((float)(2*K));
                        for (yind=1;yind<2*K;yind+=2)
                        {
                            y=((float)yind)/((float)(2*K));
                            for (zind=1;zind<2*K;zind+=2)
                            {
                                z=((float)zind)/((float)(2*K));
                                cf=x+(1.0-z)*((1.0-y)*((1.0-x)*f[i000]+x*f[i001])+y*((1.0-x)*f[i010]+x*f[i011]))+z*((1.0-y)*((1.0-x)*f[i100]+x*f[i101])+y*((1.0-x)*f[i110]+x*f[i111]));
                                cg=y+(1.0-z)*((1.0-y)*((1.0-x)*g[i000]+x*g[i001])+y*((1.0-x)*g[i010]+x*g[i011]))+z*((1.0-y)*((1.0-x)*g[i100]+x*g[i101])+y*((1.0-x)*g[i110]+x*g[i111]));
                                ce=z+(1.0-z)*((1.0-y)*((1.0-x)*e[i000]+x*e[i001])+y*((1.0-x)*e[i010]+x*e[i011]))+z*((1.0-y)*((1.0-x)*e[i100]+x*e[i101])+y*((1.0-x)*e[i110]+x*e[i111]));
                                hx=cf-floor(cf);
                                hy=cg-floor(cg);
                                hz=ce-floor(ce);
                                bj=(int)floor(cf)+j;
                                bi=(int)floor(cg)+i;
                                bk=(int)floor(ce)+k;
                                if (interpolate)
                                {
                                    cI=(1.0-z)*((1.0-y)*((1.0-x)*I[i000]+x*I[i001])+y*((1.0-x)*I[i010]+x*I[i011]))+z*((1.0-y)*((1.0-x)*I[i100]+x*I[i101])+y*((1.0-x)*I[i110]+x*I[i111]));
                                    if ((bk>=0)&&(bk<s))
                                    {
                                        if ((bj>=0)&&(bj<c))
                                        {
                                            if ((bi>=0)&&(bi<r))
                                            {
                                                cw=(1.0-hx)*(1.0-hy)*(1.0-hz);
                                                ind=rc*bk+r*bj+bi;
                                                (*WI)[ind]+=cw*cI;
                                                ws[ind]+=cw;
                                            }
                                            if ((bi>=-1)&&(bi<r-1))
                                            {
                                                cw=(1.0-hx)*hy*(1.0-hz);
                                                ind=rc*bk+r*bj+bi+1;
                                                (*WI)[ind]+=cw*cI;
                                                ws[ind]+=cw;
                                            }
                                        }
                                        if ((bj>=-1)&&(bj<c-1))
                                        {
                                            if ((bi>=0)&&(bi<r))
                                            {
                                                cw=hx*(1.0-hy)*(1.0-hz);
                                                ind=rc*bk+r*(bj+1)+bi;
                                                (*WI)[ind]+=cw*cI;
                                                ws[ind]+=cw;
                                            }
                                            if ((bi>=-1)&&(bi<r-1))
                                            {
                                                cw=hx*hy*(1.0-hz);
                                                ind=rc*bk+r*(bj+1)+bi+1;
                                                (*WI)[ind]+=cw*cI;
                                                ws[ind]+=cw;
                                            }
                                        }
                                    }
                                    if ((bk>=-1)&&(bk<s-1))
                                    {
                                        if ((bj>=0)&&(bj<c))
                                        {
                                            if ((bi>=0)&&(bi<r))
                                            {
                                                cw=(1.0-hx)*(1.0-hy)*hz;
                                                ind=rc*(bk+1)+r*bj+bi;
                                                (*WI)[ind]+=cw*cI;
                                                ws[ind]+=cw;
                                            }
                                            if ((bi>=-1)&&(bi<r-1))
                                            {
                                                cw=(1.0-hx)*hy*hz;
                                                ind=rc*(bk+1)+r*bj+bi+1;
                                                (*WI)[ind]+=cw*cI;
                                                ws[ind]+=cw;
                                            }
                                        }
                                        if ((bj>=-1)&&(bj<c-1))
                                        {
                                            if ((bi>=0)&&(bi<r))
                                            {
                                                cw=hx*(1.0-hy)*hz;
                                                ind=rc*(bk+1)+r*(bj+1)+bi;
                                                (*WI)[ind]+=cw*cI;
                                                ws[ind]+=cw;
                                            }
                                            if ((bi>=-1)&&(bi<r-1))
                                            {
                                                cw=hx*hy*hz;
                                                ind=rc*(bk+1)+r*(bj+1)+bi+1;
                                                (*WI)[ind]+=cw*cI;
                                                ws[ind]+=cw;
                                            }
                                        }
                                    }
                                }
                                else
                                {
                                    cI= ( (z<.5) ? ( (y<.5) ? ( (x<.5) ? I[i000] : I[i001] ) : ( (x<.5) ? I[i010] : I[i011] ) ) : ( (y<.5) ? ( (x<.5) ? I[i100] : I[i101] ) : ( (x<.5) ? I[i110] : I[i111] ) ) );
                                    if (hz>=.5)
                                        bk++;
                                    if (hx>=.5)
                                        bj++;
                                    if (hy>=.5)
                                        bi++;
                                    if ((bi>=0)&(bi<r)&(bj>=0)&(bj<c)&(bk>=0)&(bk<s))
                                    {
                                        ind=rc*bk+r*bj+bi;
                                        (*WI)[ind]+=cI;
                                        ws[ind]+=1.0;
                                    }
                                }
                            }
                        }
                    }
            }
    ind=0;
    for (k=0;k<s;k++)
        for (j=0;j<c;j++)
            for (i=0;i<r;i++)
            {
                if (ws[ind]>0.0) (*WI)[ind]/=ws[ind];
                ind++;
            }
    free(ws);
    return ind;
}

/********************************************************************************************************************************/	

int ReadImage3D(char *fname, float **Image, char fileformat)
{
    FILE *infile;
    unsigned char ctemp;
    unsigned short stemp;
    float ftemp;
    int cnt=0;
    
    if ((infile=fopen(fname,"rb"))==NULL)
    {
        fprintf(stderr,"Cannot open file %s for read.\n",fname);
        fprintf(stderr,"Image not read.\n");
        return 0;
    }
    switch (fileformat) 
    {
        case 'b':
            while (fread(&ctemp,sizeof(unsigned char),1,infile)>0)
            {
                (*Image)[cnt]=(float)ctemp;
                cnt++;
            }
            break;
        case 's':
            while (fread(&stemp,sizeof(unsigned short),1,infile)>0)
            {
                (*Image)[cnt]=(float)stemp;
                cnt++;
            }
            break;
        case 'f':
            while (fread(&ftemp,sizeof(float),1,infile)>0)
            {
                (*Image)[cnt]=ftemp;
                cnt++;
            }
            break;
        default:
            fprintf(stderr,"Unrecognized file format for file %s.\n",fname);
            fprintf(stderr,"Reading as a byte image.\n");
            while (fread(&ctemp,sizeof(unsigned char),1,infile)>0)
            {
                (*Image)[cnt]=(float)ctemp;
                cnt++;
            }
            break;
    }
    fclose(infile);
    
    return cnt;
}

/********************************************************************************************************************************/	

int WriteImage3D(char *fname, float *Image, char fileformat, int s)
{
    FILE *outfile;
    unsigned char ctemp;
    unsigned short stemp;
    float ftemp;
    int itemp;
    int cnt;
    
    if ((outfile=fopen(fname,"wb"))==NULL)
    {
        fprintf(stderr,"Cannot open file %s for write.\n",fname);
        fprintf(stderr,"Image not written.\n");
        return 0;
    }
    switch (fileformat) 
    {
        case 'b':
            for (cnt=0;cnt<s;cnt++)
            {
                ftemp=Image[cnt];
                itemp=(int)((ftemp-(int)ftemp >= 0.5) ? ftemp+0.5 : ftemp);
                ctemp=(unsigned char)((itemp < 0) ? itemp = 0 : ((itemp > 255) ? itemp = 255 : itemp));
                fwrite(&ctemp,sizeof(unsigned char),1,outfile);
            }
            break;
        case 's':
            for (cnt=0;cnt<s;cnt++)
            {
                ftemp=Image[cnt];
                itemp=(int)((ftemp-(int)ftemp >= 0.5) ? ftemp+0.5 : ftemp);
                stemp=(unsigned short)((itemp < 0) ? itemp = 0 : ((itemp > 65535) ? itemp = 65535 : itemp));
                fwrite(&stemp,sizeof(unsigned short),1,outfile);
            }
            break;
        case 'f':
            for (cnt=0;cnt<s;cnt++)
            {
                ftemp=Image[cnt];
                fwrite(&ftemp,sizeof(float),1,outfile);
            }
            break;
        default:
            fprintf(stderr,"Unrecognized file format for file %s.\n",fname);
            fprintf(stderr,"Writing as a byte image.\n");
            for (cnt=0;cnt<s;cnt++)
            {
                ftemp=Image[cnt];
                itemp=(int)((ftemp-(int)ftemp >= 0.5) ? ftemp+0.5 : ftemp);
                ctemp=(unsigned char)((itemp < 0) ? itemp = 0 : ((itemp > 255) ? itemp = 255 : itemp));
                fwrite(&ctemp,sizeof(unsigned char),1,outfile);
            }
            break;
    }
    fclose(outfile);
    return cnt;
}

/********************************************************************************************************************************/	

void Usage(char *programName)
{
    int i,n=0;
    char *emptyspace;
    
    while (programName[n]!='\0')
        n++;
    emptyspace=(char *)malloc(sizeof(char)*n);
    for (i=0;i<n;i++)
        emptyspace[i]=' ';
    emptyspace[n]='\0';
	fprintf(stderr,"%s usage:\n",programName);
	fprintf(stderr,"%s warps a given image according to a given deformation field. ",programName);
	fprintf(stderr,"The image and the deformation field must have equal size. ");
	fprintf(stderr,"By default, %s treats the image as a gray scale image. For volume ",programName);
	fprintf(stderr,"redistribution, -V option must be specified. For better results, large upsampling ");
	fprintf(stderr,"should be specified with -T option.\n");
    fprintf(stderr,"%s takes the following options:\n",programName);
    fprintf(stderr,"%s -I <string, input image> \n",emptyspace);
    fprintf(stderr,"%s -d <string, deformation field> \n",emptyspace);
    fprintf(stderr,"%s -r <int, number of rows (y dim) [256]> \n",emptyspace);
    fprintf(stderr,"%s -c <int, number of columns (x dim) [256]> \n",emptyspace);
    fprintf(stderr,"%s -s <int, number of slices (z dim) [198]> \n",emptyspace);
    fprintf(stderr,"%s -o <string, warped image [defaultoutput.dat]> \n",emptyspace);
    fprintf(stderr,"%s -T <int, upsampling coefficient [3]> \n",emptyspace);
    fprintf(stderr,"%s -D <for deformation fields (default is a displacement field)> \n",emptyspace);
    fprintf(stderr,"%s -V <for volume redistribution> \n",emptyspace);
    fprintf(stderr,"%s -S <for smooth interpolation> \n",emptyspace);
    fprintf(stderr,"%s -fi <b-s-f, input image format [b]> \n",emptyspace);
    fprintf(stderr,"%s -fo <b-s-f, output image format [b]> \n",emptyspace);
    
    free(emptyspace);
}

/********************************************************************************************************************************/	

/* returns the index of the first argument that is not an option; i.e.
   does not start with a dash or a slash
*/
int HandleOptions(int argc,char *argv[])
{
	int i,firstnonoption=0;
	float tau,popt;

	for (i=1; i< argc;i++) {
/*		if (argv[i][0] == '/' || argv[i][0] == '-') {*/
		if (argv[i][0] == '-') {
			switch (argv[i][1]) {
				/* An argument -? means help is requested */
				case '?':
					Usage(argv[0]);
					break;
/*				case 'h':
				case 'H':
					if (!stricmp(argv[i]+1,"help")) {
						Usage(argv[0]);
						break;
					}
					/* If the option -h means anything else
					 * in your application add code here
					 * Note: this falls through to the default
					 * to print an "unknow option" message
					*/
				/* add your option switches here */
				case 'I':
					strcpy((char *)&ImageFileName,argv[i+1]);
					break;
				case 'd':
					strcpy((char *)&DeformationFileName,argv[i+1]);
					break;
				case 'o':
					strcpy((char *)&OutputFileName,argv[i+1]);
					NOOUTFILE=0;
					break;
				case 'D':
                    DF=1;
					break;
				case 'r':
                    R=atoi(argv[i+1]);
					break;
				case 'c':
                    C=atoi(argv[i+1]);
					break;
				case 's':
                    S=atoi(argv[i+1]);
					break;
				case 'T':
                    T=atoi(argv[i+1]);
					break;
				case 'V':
                    VOLUMEDISTRIBUTION=1;
					break;
				case 'S':
                    INTERPOLATE=1;
					break;
				default:
					fprintf(stderr,"unknown option %s\n",argv[i]);
					break;
			}
		}
	}
	return 1;
}



/********************************************************************************************************************************/	

int main(int argc,char *argv[])
{
    float *Image, *f, *g, *e, *WarpedImage;
    int i,j,k,ind;
	
	if (argc == 1) {
		/* If no arguments we call the Usage routine and exit */
		Usage(argv[0]);
		return 1;
	}
	/* handle the program options */
	if (HandleOptions(argc,argv))
    {
        Image=(float *)malloc(sizeof(float)*R*C*S);
        if (ReadImage3D(ImageFileName,&Image,IF)==0)
        {
            fprintf(stderr,"Error reading file.\n");
            free(Image);
            return 0;
        }
        
        if (!(LoadDeformationFieldCartesianComponents3D(DeformationFileName,&f,&g,&e,R,C,S)))
        {
            fprintf(stderr,"deformation cannot be read from file %s\n",DeformationFileName);
            free(Image);
            return 0;
        }
        
        if (DF)
        {
            ind=0;
            for (k=1;k<=S;k++)
                for (j=1;j<=C;j++)
                    for (i=1;i<=R;i++)
                    {
                        f[ind]-=j;
                        g[ind]-=i;
                        e[ind]-=k;
                        ind++;
                    }
        }
        
        WarpedImage=(float *)malloc(sizeof(float)*R*C*S);
        
        if (VOLUMEDISTRIBUTION)
            WarpImage3D(f,g,e,Image,R,C,S,&WarpedImage,T,INTERPOLATE);
        else
            WarpGrayImage3D(f,g,e,Image,R,C,S,&WarpedImage,T,INTERPOLATE);
        
        if (NOOUTFILE)
            WriteImage3D("defaultoutput.dat",WarpedImage,OF,R*C*S);
        else
            WriteImage3D(OutputFileName,WarpedImage,OF,R*C*S);
        
        free(Image);
        free(WarpedImage);
        free(f);
        free(g);
        free(e);
        return 1;
    } else
    {
        fprintf(stderr,"no action performed.\n");
    	return 0;
    }
}

